
; ************************************************* Macros for AVR assembler programming ********************************************

; ********************************************************** Basic macros ************************************************************

; *************************************************** Release 4.2, June 10, 2020 *****************************************************

; CAUTION: Jump and Call instruction support:

; If the CPU supports only RJMP/RCALL instructions,  
; define it at the begin of the main program: #define reljump (0).
; If the CPU supports JMP/CALL insructions, omit this definition. 


; CAUTION: Supporting 16-bit writes to hardware (peripheral) registers:

; If the periphery requires the low byte to be written first (e.g., Xmega),  
; define it at the begin of the main program: #define lowbytefirst (0).
; If the periphery requires the high byte to be written first
; (e.g., ATmega or ATtiny), omit this definition. 

; CAUTION: The macros are to be assmbled togther with the common subroutines. Hence include common_subs_xx.asm,
; 			where xx is the valid version number (e.g., common_subs_0.asm). 
;			Subroutines must be included at the end of the main program

.def temp	=r19			; Register model definitions
.def al		=r24
.def ah		=r25
.def bl		=r22
.def bh		=r23
.def cl		=r20
.def ch		=r21	
.def cx3    =r2				; The 3rd byte of the extended C-reg


; **********************************************************************************************


; 				**************************** 1.  Basic Transports ****************************

.macro lit			; Load an immediate 8-bit value (literal) into a register (r0...31)
ldi temp,@1			; reg_adrs, literal
mov @0,temp
.endmacro

.macro gst			; Store a register content at a general address		
.if @0 < 0x40  		; general_adrs,reg_adrs
out @0,@1
.else 
sts @0,@1
.endif
.endmacro

.macro gwst			; Store the contents of two registers (r0...r31) as a16-bit word at a general address
#if defined(lowbytefirst)
.if @0 < 0x40		; general_adrs, reg_adrs (high byte), reg_adrs (low byte)
out @0,@2			; write low byte first
out @0+1,@1
.else 
sts @0+1,@2
sts @0,@1
.endif
#else
.if @0 < 0x40		; general_adrs, reg_adrs (high byte), reg_adrs (low byte)
out @0+1,@1			; write high byte first
out @0,@2
.else 
sts @0+1,@1
sts @0,@2
.endif
#endif
.endmacro

.macro gld			; Load a register (r0...r31) with the byte at a general address
.if @1 < 0x40		; reg_adrs,general_adrs
in @0,@1
.else 
lds @0,@1
.endif
.endmacro

.macro gwld			; Load two registers (r0...r31) with the 16-bit word at a general address
.if @2 < 0x40		; reg_adrs (high byte),reg_adrs (low byte),general_adrs
in @1,@2			; read low byte first
in @0,@2+1
.else 
lds @1,@2
lds @0,@2+1
.endif
.endmacro

.macro glit			; Move an immediate 8-bit value (literal) to a general address
ldi temp,@1			; general_adrs,literal
gst @0,temp
.endmacro

.macro gwlit		; Move an immediate 16-bit value (literal) to a general address
#if defined(lowbytefirst)
ldi temp,low(@1)	; general_adrs,litera
gst @0,temp			; write low byte first
ldi temp,high(@1)
gst @0+1,temp
#else
ldi temp,high(@1)	; general_adrs,litera
gst @0+1,temp		; write high byte first
ldi temp,low(@1)
gst @0,temp
#endif
.endmacro

.macro gmov			; Move the byte at the 2nd general address to the 1st 
gld temp,@1			; general_adrs,general_adrs
gst @0,temp
.endmacro

.macro gwmov		; Move the 16-bit word at the 2nd general address to the 1st  	
#if defined(lowbytefirst)
gld temp,@1			; general_adrs,general_adrs. Read low byte first
gst @0,temp			; store low byte first	
gld temp,@1+1		; move high byte
gst @0+1,temp		 
#else
gld temp,@1			; general_adrs,general_adrs
push temp			; read low byte first and save
gld temp,@1+1		; read high byte
gst @0+1,temp		; store high byte first
pop temp
gst @0,temp			; store low byte
#endif
.endmacro

.macro gclear		; Clear the byte at a general address
eor temp,temp		; general_adrs
gst @0,temp
.endmacro

.macro gwclear		; Clear the 16-bit word at a general address
#if defined (lowbytefirst)
eor temp,temp		; general_adrs
gst @0,temp			; store low byte first
gst @0+1,temp
#else
eor temp,temp		; general_adrs
gst @0+1,temp		; store high byte first
gst @0,temp
#endif
.endmacro

.macro gzero		; Test, whether the byte at a general address is zero
gld temp,@0			; general_adrs
and temp,temp
.endmacro

.macro gwzero		; Test, whether the 16-bit word at a general address is zero
gld temp,@0			; general_adrs
and temp,temp		; read low byte first
brne e1
gld temp,@0+1
and temp,temp
e1:
.endmacro

.macro zero			; Test, whether the register content (r0...31) is zero
and @0,@0			; register_adrs
.endm

; **********************************************************************************************

; 					2.  Basic Single-Bit Operations ****************************

;					**************** Clear a single bit **************** 

.macro gbit0		; Clear a bit at a general adrs
.if @0 < 0x20 		; general_adrs,bit_adrs
cbi @0,@1
.else  
gld temp,@0
andi temp, ~(1<<@1)
gst @0,temp
.endif
.endmacro

.macro ubit0		; Clear a bit in an unified bit-field
.if (@0 >>  8) < 0x20 		; unified_bitfield_adrs
cbi (@0 >> 8),(@0 & 0x00ff)
.else  
gld temp,(@0 >>  8)
andi temp, ~(1<< (@0 & 0x00ff))
gst (@0 >>  8),temp
.endif
.endmacro

.macro bit0			; Clear a bit in a register r0...r31		
ldi temp, ~(1<<@1)	; register adrs, bit adrs
and @0,temp 
.endmacro

.macro bitf0		; Clear bit in a register r16 to r31
andi @0, ~(1<<@1)	; register adrs, bit adrs
.endmacro

.macro dbitx0		; Clear a bit in memory (SRAM). Addressed via the X-register + displacement
pushy				; save the Y-register. The X address rgister does not support displacements
movw yl,xl			; X to Y
ldd temp,y+@0
andi temp, ~(1<<@1)
std y+@0,temp
popy				; restore Y
.endmacro

.macro dbity0		; Clear a bit in memory (SRAM). Addressed via the Y-register + displacement
ldd temp,y+@0
andi temp, ~(1<<@1)
std y+@0,temp
.endmacro

.macro dbitz0		; Clear a bit in memory (SRAM). Addressed via the Z-register + displacement
ldd temp,z+@0
andi temp, ~(1<<@1)
std z+@0,temp
.endmacro


;					**************** Set a single bit bit **************** 


.macro gbit1		; Set a bit at a general adrs
.if @0 < 0x20 		; general_adrs,bit_adrs
sbi @0,@1
.else  
gld temp,@0
ori temp, 1<<@1
gst @0,temp
.endif
.endmacro

.macro ubit1		; Set a bit in an unified bit-field
.if (@0 >>  8) < 0x20 		; unified_bitfield_adrs
sbi (@0 >> 8),(@0 & 0x00ff)
.else  
gld temp,(@0 >>  8)
ori temp, (1<< (@0 & 0x00ff))
gst (@0 >>  8),temp
.endif
.endmacro

.macro bit1			; Set bit in a register r0...r31		
ldi temp, 1<<@1		; register adrs, bit adrs
or @0,temp
.endmacro

.macro bitf1		; Set bit in a register r16...r31		
ori @0, 1<<@1		; register adrs, bit adrs
.endmacro

.macro dbitx1		; Set a bit in memory (SRAM). Addressed via the X-register + displacement
pushy				; save the Y-register. The X address rgister does not support displacements
movw yl,xl			; X to Y
ldd temp,y+@0
ori temp, ~(1<<@1)
std y+@0,temp
popy				; restore Y
.endmacro

.macro dbity1		; Clear a bit in memory (SRAM). Addressed via the Y-register + displacement
ldd temp,y+@0
ori temp, 1<<@1
std y+@0,temp
.endmacro


.macro dbitz1		; Clear a bit in memory (SRAM). Addressed via the Z-register + displacement
ldd temp,z+@0
ori temp, 1<<@1
std z+@0,temp
.endmacro


;					**************** Toggle a single bit **************** 

.macro gbitt		; Toggle a bit at a general adrs
gld temp,@0
sbrc temp,@1
rjmp e1				; bit=1; set => 0
ori temp, 1<<@1		; bit=0; set => 1 
rjmp e2
e1:
andi temp, ~(1<<@1)
e2:
gst @0,temp
.endmacro

.macro ubitt		; Toggle a bit in an unified bit-field
gld temp,(@0 >>8)
sbrc temp,(@0 & 0x00ff)
rjmp e1				; bit=1; set => 0
ori temp, 1<< (@0 & 0x00ff)		; bit=0; set => 1 
rjmp e2
e1:
andi temp, ~(1<< (@0 & 0x00ff))
e2:
gst (@0 >> 8),temp
.endmacro

.macro bitt			; Toggle a bit in a register r0...r31		
ldi temp, 1<<@1		; register_adrs,bit_adrs
eor @0,temp
.endmacro


.macro dbitxt		; Toggle a bit in memory (SRAM). Addressed via the X-register + displacement
pushy				; save the Y-register. The X address rgister does not support displacements
movw yl,xl			; X to Y
ldd temp,y+@0
sbrc temp,@1
rjmp e1				; bit=1; set => 0
ori temp, 1<<@1		; bit=0; set => 1 
rjmp e2
e1:
andi temp, ~(1<<@1)
e2:
std y+@0,temp
popy				; restore Y
.endmacro

.macro dbityt		; Toggle a bit in memory (SRAM). Addressed via the Y-register + displacement
ldd temp,y+@0
sbrc temp,@1
rjmp e1				; bit=1; set => 0
ori temp, 1<<@1		; bit=0; set => 1 
rjmp e2
e1:
andi temp, ~(1<<@1)
e2:
std y+@0,temp
.endmacro

.macro dbitzt		; Toggle a bit in memory (SRAM). Addressed via the Z-register + displacement
ldd temp,z+@0
sbrc temp,@1
rjmp e1				; bit=1; set => 0
ori temp, 1<<@1		; bit=0; set => 1 
rjmp e2
e1:
andi temp, ~(1<<@1)
e2:
std z+@0,temp
.endmacro


;					**************** Move a flag (ZF or CF) into a single bit **************** 

.macro gbitz		; Set a bit at a general address according to the zero flag (ZF)
brne e1				; general_adrs,bit_adrs
gbit0 @0,@1			; ZF=1 (zero). Clear bit
rjmp e2
e1:
gbit1 @0,@1			; ZF=0 (not zero). Set bit
e2:
.endmacro

.macro ubitz		; Set a bit in an unified-bitfield according to the zero flag (ZF)
brne e1				; general_adrs,bit_adrs
gbit0 (@0>>8),(@0 &0x00ff)	; ZF=1 (zero). Clear bit
rjmp e2
e1:
gbit1 (@0>>8),(@0 &0x00ff)	; ZF=0 (not zero). Set bit
e2:
.endmacro

.macro 	bitz		; Set a bit in a register (r0...r31) according to the zero flag (ZF)
brne e1				; register_adrs,bit_adrs
bit0 @0,@1			; ZF=1 (zero). Clear bit
rjmp e2
e1:
bit1 @0,@1			; ZF=0 (not zero). Set bit
e2:
.endmacro

.macro dbitxz		; Set a bit in memory (SRAM) according to the zero flag (ZF) 
					; Addressed via the X-register + displacement
pushy				; save the Y-register. The X address rgister does not support displacements
movw yl,xl			; X to Y
ldd temp,y+@0
brne e1				; general_adrs,bit_adrs
andi temp,~(1<<@1)	; ZF=1 (zero). Clear bit
rjmp e2
e1:
ori temp,1<<@1		; ZF=0 (not zero). Set bit
e2:
std yx+@0,temp
popy				; restore Y
.endmacro

.macro dbityz		; Set a bit in memory (SRAM) according to the zero flag (ZF)
					; Addressed via the Y-register + displacement
ldd temp,y+@0
brne e1				; general_adrs,bit_adrs
andi temp,~(1<<@1)	; ZF=1 (zero). Clear bit
rjmp e2
e1:
ori temp,1<<@1		; ZF=0 (not zero). Set bit
e2:
std y+@0,temp
.endmacro

.macro dbitzz		; Set a bit in memory (SRAM) according to the zero flag (ZF)
					; Addressed via the X-register + displacement
ldd temp,z+@0
brne e1				; general_adrs,bit_adrs
andi temp,~(1<<@1)	; ZF=1 (zero). Clear bit
rjmp e2
e1:
ori temp,1<<@1		; ZF=0 (not zero). Set bit
e2:
std z+@0,temp
.endmacro

.macro gbitc		; Set a bit at a general address according to the carry flag (CF)
brcs e1				; general_adrs,bit_adrs
gbit0 @0,@1			; CF=0 (carry cleared). Clear bit
rjmp e2
e1:
gbit1 @0,@1			; CF=1 (carry set). Set bit
e2:
.endm

.macro ubitc		; Set a bit in an unified bit-field according to the carry flag (CF)
brcs e1				; general_adrs,bit_adrs
gbit0 (@0>>8),(@0 &0x00ff)			; CF=0 (carry cleared). Clear bit
rjmp e2
e1:
gbit1 (@0>>8),(@0 &0x00ff)			; CF=1 (carry set). Set bit
e2:
.endmacro

.macro ubitci		; Set a bit in an unified bit-field according to the inverted carry flag (CF)
brcc e1				; general_adrs,bit_adrs
gbit0 (@0>>8),(@0 &0x00ff)			; CF=1 (carry set). Clear bit
rjmp e2
e1:
gbit1 (@0>>8),(@0 &0x00ff)			; CF=0 (carry cleared). Set bit
e2:
.endmacro

.macro 	bitc		; Set a bit in a register (r0...r31) according to the carry flag (CF)
brcs e1				; register_adrs,bit_adrs
bit0 @0,@1			; CF=0 (carry cleared). Clear bit
rjmp e2
e1:
bit1 @0,@1			; CF=1 (carry set). Set bit
e2:
.endmacro

.macro dbitxc		; Set a bit in memory (SRAM)  according to the carry flag (CF)
					; Addressed via the X-register + displacement
pushy				; save the Y-register. The X address rgister does not support displacements
movw yl,xl			; X to Y
ldd temp,x+@0
brcs e1				; general_adrs,bit_adrs
andi temp,~(1<<@1)	; ZF=1 (zero). Clear bit
rjmp e2
e1:
ori temp,1<<@1		; ZF=0 (not zero). Set bit
e2:
std x+@0,temp
popy				; restore Y
.endmacro

.macro dbityc		; Set a bit in memory (SRAM)  according to the carry flag (CF)
					; Addressed via the Y-register + displacement
ldd temp,y+@0
brcs e1				; general_adrs,bit_adrs
andi temp,~(1<<@1)	; ZF=1 (zero). Clear bit
rjmp e2
e1:
ori temp,1<<@1		; ZF=0 (not zero). Set bit
e2:
std y+@0,temp
.endmacro

.macro dbitzc		; Set a bit in memory (SRAM)  according to the carry flag (CF)
					; Addressed via the Z-register + displacement
ldd temp,z+@0
brcs e1				; general_adrs,bit_adrs
andi temp,~(1<<@1)	; ZF=1 (zero). Clear bit
rjmp e2
e1:
ori temp,1<<@1		; ZF=0 (not zero). Set bit
e2:
std z+@0,temp
.endmacro


;					**************** Test (query, examine) a single bit **************** 

.macro gbit			; Test a bit at a general address for zero		
gld temp,@0			; general_adrs,bit_adrs
andi temp, 1<<@1
.endmacro

.macro ubit			; Test a bit in an unified bit-field for zero		
gld temp,(@0>>8)			
andi temp, 1<<(@0 & 0x00ff)
.endmacro

.macro bit			; Test a bit in a register (r0...r31) for zero	
mov temp,@0			; register_adrs,bit_adrs
andi temp, 1<<@1
.endmacro

.macro dbitx		; Test a bit in memory (SRAM)for zero
					; Addressed via the X-register + displacement
pushy				; save the Y-register. The X address rgister does not support displacements
movw yl,xl			; X to Y
ldd temp,y+@0
andi temp, ~(1<<@1)
popy				; restore Y
.endmacro

.macro dbity		; Test a bit in memory (SRAM)for zero
					; Addressed via the Y-register + displacement
ldd temp,y+@0
andi temp, ~(1<<@1)
.endmacro

.macro dbitz		; Test a bit in memory (SRAM) for zero
					; Addressed via the Z-register + displacement
ldd temp,z+@0
andi temp, ~(1<<@1)
.endmacro

.macro gbittoc		; Move a bit at a general address into the carry flag (CF)		
gld temp,@0			; general_adrs,bit_adrs
clc					; clear CF
sbrc temp,@1		; Bit set?
sec					; Yes. Set CF
.endmacro

.macro ubittoc		; Move a bit in an unified bit-field into the carry flag (CF)	
gld temp,(@0>>8)			 
clc							; clear CF
sbrc temp,(@0 & 0x00ff)		; Bit set?
sec							; Yes. Set CF
.endmacro

.macro bittoc		; Move a bit in a register (r0...r31) into the carry flag (CF)	
mov temp,@0			; register_adrs,bit_adrs
clc					; clear CF
sbrc temp,@1		; Bit set?
sec					; Yes. Set CF
.endmacro

.macro dbitxtoc		; Move a bit  in memory (SRAM) into the carry flag (CF)
					; Addressed via the X-register + displacement
pushy				; save the Y-register. The X address rgister does not support displacements
movw yl,xl			; X to Y
ldd temp,y+@0
clc					; clear CF
sbrc temp,@1		; Bit set?
sec					; Yes. Set CF
popy				; restore Y
.endmacro

.macro dbitytoc		; Move a bit  in memory (SRAM) into the carry flag (CF)
					; Addressed via the Y-register + displacement
ldd temp,y+@0
clc					; clear CF
sbrc temp,@1		; Bit set?
sec					; Yes. Set CF
.endmacro


.macro dbitztoc		; Move a bit  in memory (SRAM) into the carry flag (CF)
					; Addressed via the X-register + displacement
ldd temp,z+@0
clc					; clear CF
sbrc temp,@1		; Bit set?
sec					; Yes. Set CF
.endmacro


; **********************************************************************************************

; 					3.  Jump on a flag ****************************


.macro brz			; Short branch on zero (other name for BREQ)	
breq @0				; branch_adrs
.endmacro

.macro brnz			; Short branch on not zero (other name for BRNE)
brne @0				; branch_adrs
.endmacro

.macro jp			; Jump always (unconditional)
#if defined(reljump)
rjmp @0
#else 
jmp @0
#endif
.endmacro

.macro goto			; Jump always (unconditional). Other name for JP
jp @0
.endmacro

.macro jpz			; Jump if zero
brne e1
jp @0
e1:
.endmacro

.macro jpeq			; Jump if equal. Other name for JPZ
jpz @0
.endmacro

.macro jpnz			; Jump if not zero
breq e1
jp @0
e1:
.endmacro

.macro jpne			; Jump if not equal. Other name for JPNZ
jpnz @0
.endmacro

.macro jpc			; Jump if carry (CF = 1)
brcc e1
jp @0			 
e1:
.endmacro

.macro jpnc			; Jump if no carry (CF = 0)
brcs e1
jp @0			 
e1:
.endmacro

.macro jpn			; Jump if negative (SF = 1)
brpl e1
jp @0			 
e1:
.endmacro

.macro jpp			; Jump if positive (SF = 0)
brmi e1
jp @0			 
e1:
.endmacro

.macro jpov			; Jump if overflow (OF = 1) Two's complement overflow
brvc e1
jp @0			 
e1:
.endmacro

.macro jpnov		; Jump if no overflow (OF = 0)
brvs e1
jp @0			 
e1:
.endmacro

.macro jplf			; Jump if less (LF = 1). A less than B as the result of A - B)
brge e1
jp @0			 
e1:
.endmacro

.macro jpnlf		; Jump if not less (LF = 0). A greater or equal B
brlt e1
jp @0			 
e1:
.endmacro

.macro jph			; Jump if carry-out of low-order 4-bit nibble (HF = 1). Half-byte carry
brhc e1
jp @0			 
e1:
.endmacro

.macro jpnh			; Jump if no carry-out of low-order 4-bit nibble (HF = 0)
brhs e1
jp @0			 
e1:
.endmacro

.macro jpt			; Jump if  transfer-bit set (TF = 1)
brtc e1
jp @0			 
e1:
.endmacro

.macro jpnt			; Jump if  transfer-bit cleared (TF = 0) 
brts e1
jp @0			 
e1:
.endmacro


.macro jen			; Jump if interrups enabled (IF = 1)
brid e1
jp @0			 
e1:
.endmacro


.macro jdi			; Jump if interrups disabled (IF = 0)
brie e1
jp @0			 
e1:
.endmacro


; **********************************************************************************************

; 					4.  Jump on a single bit ****************************


.macro jpg0			; Jump if a bit at general adrs is = 0
gbit @0,@1			; general_adrs,bit_adrs,jump adrs
jpz @2		
.endmacro

.macro jpg1			; Jump if a bit at general adrs is = 1
gbit @0,@1			; general_adrs,bit_adrs,jump_adrs
jpnz @2
.endmacro

.macro jp0			; Jump if a bit in a register (r0...r31) is = 0
bit @0,@1			; register_adrs,bit_adrs,jump_adrs
jpz @2
.endmacro


.macro jp1			; Jump if a bit in a register (r0...r31) is = 1
bit @0,@1			; register_adrs,bit_adrs,jump_adrs
jpnz @2				
.endmacro

.macro jpu0			; Jump if a bit in an unified bit-field is = 0
ubit @0
jpz @1
.endmacro

.macro jpu1			; Jump if a bit in an unified bit-field is = 1
ubit @0
jpnz @1
.endmacro

.macro jpx0			; Jump if a bit in memory (SRAM) is = 0
					; Addressed via the X-register + displacement
dbitx @0			
jpz @1
.endmacro

.macro jpx1			; Jump if a bit in memory (SRAM) is = 1
					; Addressed via the X-register + displacement
dbitx @0			
jpnz @1
.endmacro

.macro jpy0			; Jump if a bit in memory (SRAM) is = 0
					; Addressed via the Y-register + displacement
dbity @0			
jpz @1
.endmacro

.macro jpy1			; Jump if a bit in memory (SRAM) is = 1
					; Addressed via the Y-register + displacement
dbity @0			
jpnz @1
.endmacro

.macro jpz0			; Jump if a bit in memory (SRAM) is = 0
					; Addressed via the Z-register + displacement
dbitz @0			
jpz @1
.endmacro

.macro jpz1			; Jump if a bit in memory (SRAM) is = 1
					; Addressed via the Z-register + displacement
dbitz @0			
jpnz @1
.endmacro


; **********************************************************************************************

; 					5.  Jump after subtracting A - B (compare arithmetically) ******************


.macro jpb			; Jump if below (A < B; unsigned)
brcc e1
jp @0				; jump if CF set 
e1:
.endmacro

.macro jpbe			; Jump if below or equal (A <= B; unsigned)
brcs e1				; jump if CF set (below)
brne e2				; jump if ZF set (equal); do not jump if ZF cleared
e1:
jp @0				
e2:
.endmacro

.macro jpa			; Jump if above (A > B; unsigned)
brcs e1				; do not jump if CF set
breq e1				; do not jump if ZF set
jp @0				; jump if CF cleared and ZF cleared
e1:
.endmacro

.macro jpae			; Jump if above or equal (A >= B; unsigned)
brcs e1				; do not jump if CF set
jp @0				; jump if CF cleared
e1:
.endmacro

.macro jpl			; Jump if less  (A < B; signed)
brge e1				; do not jump if SF (NEG xor OV) cleared
jp @0				; jump if SF set 
e1:
.endmacro

.macro jple			; Jump if less or equal  (A <= B; signed)
brlt e1				; jump if SF set (less)
brne e2				; jump if ZF set (equal); do not jump if ZF cleared
e1:
jp @0				 
e2:
.endmacro

.macro jpg			; Jump if greater  (A > B; signed)
breq e1				; do not jump if ZF set (equal)
brlt e1				; do not jump if SF set (less)
jp @0
e1:
.endmacro

.macro jpge			; Jump if greater or equal  (A >= B; signed)
brlt e1				; do not jump if SF set (less)
jp @0
e1:
.endmacro


; **********************************************************************************************

; 					6.  Skip a single instruction word ****************************

;						Arithmetic conditions are related to the result of subtracting A - B

.macro skipz		; Skip if zero 		
breq PC+2
.endmacro

.macro skipeq		; Skip if equal. Oher name for SKIPZ
breq PC+2
.endmacro

.macro skipnz		; Skip if not zero 
brne PC+2
.endmacro

.macro skipne		; Skip if not equal. Other name for SKIPNZ
brne PC+2
.endmacro

.macro skipb		; Skip if below (A < B; unsigned)
brcs PC+2			; skip if CF set
.endmacro

.macro skipbe		; Skip if below or equal (A <= B; unsigned)
brcs e1+2			; skip if CF set (below)
breq e1+2			; skip if ZF set (equal)
e1:
.endmacro

.macro skipa		; Skip if above (A > B; unsigned)
brcs e1				; do not skip if CF set
breq e1				; do not skip if ZF set
rjmp e1+2			; skip if CF cleared and ZF cleared
e1:
.endmacro

.macro skipae		; Skip if above or equal (A >= B; unsigned)
brcc PC+2			; skip if CF cleared
.endmacro

.macro skipl		; Skip if less  (A < B; signed)
brlt PC+2			; skip if SF (NEG xor OV) set
.endmacro

.macro skiple		; Skip if less or equal  (A <= B; signed)
brlt e1+2			; skip if SF (NEG xor OV) set (less)
breq e1+2			; skip if ZF set (equal)
e1:
.endmacro

.macro skipg		; Skip if greater  (A > B; signed)
breq e1				; do not skip if ZF set (equal)
brlt e1				; do not skip if SF set (less)
rjmp e1+2
e1:
.endmacro

.macro skipge		; Skip if greater or equal  (A >= B; signed)
brge PC+2			; skip if SF cleared (greater or equal)
.endmacro

.macro skipg0		; Skip if a bit at a general address is = 0
gbit @0,@1			; general_adrs,bit_adrs	
breq PC+2
.endmacro

.macro skipg1		; Skip if a bit at a general address is = 1
gbit @0,@1			; general_adrs,bit_adrs
brne PC+2
.endmacro

.macro skipu0		; Skip if a bit in an unified bit-field is = 0
ubit @0			
breq PC+2
.endmacro

.macro skipu1		; Skip if a bit in an unified bit-field is = 1
ubit @0			
brne PC+2
.endmacro


; **********************************************************************************************

; 					7.  Skip two instruction words ****************************

;						Arithmetic conditions are related to the result of A - B

.macro skipz2		; Skip if zero 		
breq PC+3
.endmacro

.macro skipeq2		; Skip if equal. Other name for SKIPZ2 
breq PC+3
.endmacro

.macro skipnz2		; Skip if not zero 
brne PC+3
.endmacro

.macro skipne2		; Skip if not equal. Other name for SKIPNZ2 
brne PC+3
.endmacro

.macro skipb2		; Skip if below (A < B; unsigned)
brcs PC+3			; skip if CF set
.endmacro

.macro skipbe2		; Skip if below or equal (A <= B; unsigned)
brcs e1+3			; skip if CF set (below)
breq e1+3			; skip if ZF set (equal)
e1:
.endmacro

.macro skipa2		; Skip if above (A > B; unsigned)
brcs e1				; do not skip if CF set
breq e1				; do not skip if ZF set
rjmp e1+3			; skip if CF cleared and ZF cleared
e1:
.endmacro

.macro skipae2		; Skip if above or equal (A >= B; unsigned)
brcc PC+3			; skip if CF cleared
.endmacro

.macro skipl2		; Skip if less  (A < B; signed)
brlt PC+3			; skip if SF (NEG xor OV) set
.endmacro


.macro skiple2		; Skip if less or equal  (A <= B; signed)
brlt e1+3			; skip if SF (NEG xor OV) set (less)
breq e1+3			; skip if ZF set (equal)
e1:
.endmacro

.macro skipg2		; Skip if greater  (A > B; signed)
breq e1				; do not skip if ZF set (equal)
brlt e1				; do not skip if SF set (less)
rjmp e1+3
e1:
.endmacro

.macro skipge2		; Skip if greater or equal  (A >= B; signed)
brge PC+3			; skip if SF cleared (greater or equal)
.endmacro

.macro skipg02		; Skip if a bit at a general address is = 0
gbit @0,@1			; general_adrs,bit_adrs	
breq PC+3
.endmacro

.macro skipg12		; Skip if a bit at a general address is = 1
gbit @0,@1			; general_adrs,bit_adrs
brne PC+3
.endmacro

.macro skipu02		; Skip if a bit in an unified bit-field is = 0
ubit @0			
breq PC+3
.endmacro

.macro skipu12		; Skip if a bit in an unified bit-field is = 1
ubit @0			
brne PC+3
.endmacro


; **********************************************************************************************

; 					8.  Skip  the following instruction ****************************

; 						The number of instruction words to be skipped will be considered automatically.

;						Arithmetic conditions are related to the result of A - B


.macro skipfz		; Skip if zero (ZF = 1)	
in temp,sreg
sbrs temp,1
.endmacro

.macro skipfeq		; Skip if equal. Other name for SKIPFZ 
in temp,sreg
sbrs temp,1
.endmacro

.macro skipfnz		; Skip if not zero (ZF = 0)	
in temp,sreg
sbrc temp,1
.endmacro

.macro skipfne		; Skip if not equal. Other name for SKIPFNZ
in temp,sreg
sbrc temp,1
.endmacro

.macro skipfc		; Skip if carry (CF = 1)		
in temp,sreg
sbrs temp,0
.endmacro

.macro skipfnc		; Skip if no carry 	(CF = 0)	
in temp,sreg
sbrc temp,0
.endmacro

.macro skipfn		; Skip if negative 	(SF = 1)	
in temp,sreg
sbrs temp,2
.endmacro

.macro skipfp		; Skip if positive 	(SF = 0)	
in temp,sreg
sbrc temp,2
.endmacro

.macro skipfov		; Skip if overflow	(OF = 1)	
in temp,sreg
sbrs temp,3
.endmacro

.macro skipfnov		; Skip if no overflow (OF = 0)	
in temp,sreg
sbrc temp,3
.endmacro

.macro skipflf		; Skip if less (A < B; signed)	(LF = 1)
in temp,sreg		
sbrs temp,4
.endmacro

.macro skipfnlf		; Skip if not less = greater or equal. A >= B; signed)	(LF = 0)
in temp,sreg		 
sbrc temp,4
.endmacro

.macro skipfh		; Skip if carry-out of low-order 4-bit nibble (half-byte carry)	(HF = 1)	
in temp,sreg
sbrs temp,5
.endmacro

.macro skipfnh		; Skip if  no carry-out of low-order 4-bit nibble	(HF = 0)	
in temp,sreg
sbrc temp,5
.endmacro

.macro skipft		; Skip if transfer-bit set 	(TF = 1)	
in temp,sreg
sbrs temp,6
.endmacro

.macro skipfnt		; Skip if transfer-bit cleared	(TF = 0)
in temp,sreg
sbrc temp,6
.endmacro

.macro skipfen		; Skip if interrupts enabled 	(IF = 1)	
in temp,sreg
sbrs temp,7
.endmacro

.macro skipfdi		; Skip if interrupts disabled	(IF = 0)	
in temp,sreg
sbrc temp,7
.endmacro

.macro skipfb		; Skip if below (A < B; unsigned)
in temp,sreg
sbrs temp,0			; skip if CF set
.endmacro

.macro skipfbe		; Skip if below or equal (A <= B; unsigned)
brcs e1				; skip if CF set (below)
brne e2 			; skip if ZF set (equal). Do not skip if ZF cleared
e1:
eor temp,temp		; clear temp
sbrc temp,0			; skip unconditionally (because bit is zero)
e2:
.endmacro

.macro skipfa		; Skip if above (A > B; unsigned)
brcs e1				; do not skip if CF set
breq e1				; do not skip if ZF set
eor temp,temp		; clear temp
sbrc temp,0			; skip unconditionally (because bit is zero)
e1:
.endmacro

.macro skipfae		; Skip if above or equal (A >= B; unsigned)
in temp,sreg
sbrc temp,0			; skip if CF cleared
.endmacro

.macro skipfl		; Skip if less  (A < B; signed)
in temp,sreg
sbrs temp,4			; skip if LF (NEG xor OV) set
.endmacro

.macro skipfle		; Skip if less or equal  (A <= B; signed)
brlt e1				; skip if SF (NEG xor OV) set (less)
brne e2				; skip if ZF set (equal). Do not skip if ZF cleared
e1:
eor temp,temp		; clear temp
sbrc temp,0			; skip unconditionally (because bit is zero)			; 
e2:
.endmacro

.macro skipfg		; Skip if greater  (A > B; signed)
breq e1				; do not skip if ZF set (equal)
brlt e1				; do not skip if SF set (less)
eor temp,temp		; clear temp
sbrc temp,0			; skip unconditionally (because bit is zero)
e1:
.endmacro

.macro skipfge		; Skip if greater or equal  (A >= B; signed)
in temp,sreg
sbrc temp,4			; skip if LF cleared (greater or equal)
.endmacro

.macro skipfg0		; Skip if a bit at a general address is = 0
gld temp, @0		; load from general adrs
sbrc temp,@1		; skip on bit adrs	
.endmacro

.macro skipfg1		; Skip if a bit at a general address is = 1
gld temp, @0		; load from general adrs
sbrs temp,@1		; skip on bit adrs	
.endmacro

.macro skipfu0		; Skip if a bit in an unified bit-field is = 0
gld temp,(@0>>8)			
sbrc temp,(@0 & 0x00ff)
.endmacro

.macro skipfu1		; Skip if a bit in an unified bit-field is = 1
gld temp,(@0>>8)			
sbrs temp,(@0 & 0x00ff)
.endmacro


; **********************************************************************************************

; 					9.  Call a subroutine ****************************

;						Arithmetic conditions are related to the result of subtracting A - B


.macro cal			; Call always (unconditional)
#if defined(reljump)
rcall @0
#else 
call @0
#endif
.endmacro

.macro gosub		; Call always (unconditional). Other name for CAL
cal @0
.endmacro

.macro callz		; Call if zero
brne e1
cal @0
e1:
.endmacro

.macro calleq		; Call if equal. Other name for CALLZ
callz @0
.endmacro

.macro callnz		; Call if not zero
breq e1
cal @0
e1:
.endmacro

.macro callne		; Call if not equal. Other name for CALLNZ
callnz @0
.endmacro

.macro callb		; Call if below (A < B; unsigned)
brcc e1
cal @0				; call if CF set 
e1:
.endmacro

.macro callbe		; Call if below or equal (A <= B; unsigned)
brcs e1				; call if CF set (below)
brne e2				; call if ZF set (equal); do not call if ZF cleared 
e1:
call @0
e2:					
.endmacro

.macro calla		; Call if above (A > B; unsigned)
brcs e1				; do not call if CF set
breq e1				; do not call if ZF set
cal @0				; call if CF cleared and ZF cleared
e1:
.endmacro

.macro callae		; Call if above or equal (A >= B; unsigned)
brcs e1				; do not call if CF set
cal @0				; call if CF cleared
e1:
.endmacro

.macro calll		; Call if less  (A < B; signed)
brge e1				; do not call if SF (NEG xor OV) cleared
cal @0				; call if SF set 
e1:
.endmacro

.macro callle		; Call if less or equal  (A <= B; signed)
brlt e1				; call if SF set (less)
brne e2				; call if ZF set (equal); do not call if ZF cleared
e1:
cal @0				
e2:
.endmacro

.macro callg		; Call if greater  (A > B; signed)
breq e1				; do not call if ZF set (equal)
brlt e1				; do not call if SF set (less)
cal @0
e1:
.endmacro

.macro callge		; Call if greater or equal  (A >= B; signed)
brlt e1				; do not call if SF set (less)
cal @0
e1:
.endmacro

.macro callg0		; Call if a bit at a general address is = 0
gbit @0,@1			; general_adrs,bit_adrs,subroutine_adrs
callz @2		
.endmacro

.macro callg1		; Call if a bit at a general address is = 1
gbit @0,@1			; general_adrs,bit_adrs,subroutine_adrs
callnz @2
.endmacro

.macro call0		; Call if a bit in a register (r0...r31) is = 0
bit @0,@1			; register_adrs,bit_adrs,subroutine_adrs
callz @2
.endmacro

.macro call1		; Call if a bit in a register (r0...r31) is = 1
bit @0,@1			; register_adrs,bit_adrs,subroutine_adrs
callnz @2				
.endmacro

.macro callu0		; Call if a bit in a unified bit-field is = 0
ubit @0
callz @1
.endmacro

.macro callu1		; Call if a bit in a unified bit-field is = 1
ubit @0
callnz @1
.endmacro

.macro callx0		; Call if a bit in memory (SRAM) is = 0
					; Addressed via the X-register + displacement
dbitx @0			
callz @1
.endmacro

.macro callx1		; Call if a bit in memory (SRAM) is = 1
					; Addressed via the X-register + displacement
dbitx @0			
callnz @1
.endmacro

.macro cally0		; Call if a bit in memory (SRAM) is = 0
					; Addressed via the Y-register + displacement
dbity @0			
callz @1
.endmacro

.macro cally1		; Call if a bit in memory (SRAM) is = 1
					; Addressed via the Y-register + displacement
dbity @0			
callnz @1
.endmacro

.macro callz0		; Call if a bit in memory (SRAM) is = 0
					; Addressed via the Z-register + displacement
dbitz @0			
callz @1
.endmacro

.macro callz1		; Call if a bit in memory (SRAM) is = 1
					; Addressed via the Z-register + displacement
dbitz @0			
callnz @1
.endmacro


; **********************************************************************************************

; 					10.  Return from a subroutine ****************************

;						Arithmetic conditions are related to the result of subtracting A - B


.macro retz			; Return if zero
brne PC+2
ret 
.endmacro

.macro reteq		; Return if equal. Other name for RETZ
retz 
.endmacro

.macro retnz		; Return if not zero
breq PC+2
ret
.endmacro

.macro retne		; Return if not equal: Other name for RETNZ
retnz 
.endmacro

.macro retb			; Return if below (A < B; unsigned)
brcc PC+2
ret					; return if CF set 
e1:
.endmacro

.macro retbe		; Return if below or equal (A <= B; unsigned)
brcs e1				; return if CF set (below)
brne e2				; return if ZF set (equal); do not return if ZF cleared 
e1:
ret
e2:					
.endmacro

.macro reta			; Return if above (A > B; unsigned)
brcs e1				; do not return if CF set
breq e1				; do not return if ZF set
ret					; return if CF cleared and ZF cleared
e1:
.endmacro

.macro retae		; Return if above or equal (A >= B; unsigned)
brcs PC+2			; do not return if CF set
ret					; return if CF cleared
.endmacro

.macro retl			; Return if less  (A < B; signed)
brge PC+2			; do not return if SF (NEG xor OV) cleared
ret					; return if SF set 
.endmacro

.macro retle		; Return if less or equal  (A <= B; signed)
brlt e1				; return if SF set (less)
brne e2				; return if ZF set (equal); do not return if ZF cleared
e1:
ret				
e2:
.endmacro

.macro retg			; Return if greater  (A > B; signed)
breq e1				; do not return if ZF set (equal)
brlt e1				; do not return if SF set (less)
ret
e1:
.endmacro

.macro retge		; Return if greater or equal  (A >= B; signed)
brlt PC+2			; do not return if SF set (less)
ret
.endmacro

.macro retg0		; Return if a bit at a general address is = 0
gbit @0,@1			; general_adrs,bit_adrs
retz		
.endmacro

.macro retg1		; Return if a bit at a general address is = 1
gbit @0,@1			; general_adrs,bit_adrs
retnz 
.endmacro

.macro ret0			; Return if a bit in a register (r0...r31) is = 0
bit @0,@1			; register_adrs,bit_adrs
retz
.endmacro

.macro ret1			; Return if a bit in a register (r0...r31) is = 1
bit @0,@1			; register_adrs,bit_adrs
retnz 				
.endmacro

.macro retu0		; Return if a bit in a unified bit-field is = 0
ubit @0
retz 
.endmacro

.macro retu1		; Return if a bit in a unified bit-field is = 1
ubit @0
retnz 
.endmacro

.macro retx0		; Return if a bit in memory (SRAM) is = 0
					; Addressed via the X-register + displacement
dbitx @0			
retz 
.endmacro

.macro retx1		; Return if a bit in memory (SRAM) is = 1
					; Addressed via the X-register + displacement
dbitx @0			
retnz 
.endmacro

.macro rety0		; Return if a bit in memory (SRAM) is = 0
					; Addressed via the Y-register + displacement
dbity @0			
retz
.endmacro

.macro rety1		; Return if a bit in memory (SRAM) is = 1
					; Addressed via the Y-register + displacement
dbity @0			
retnz
.endmacro

.macro retz0		; Return if a bit in memory (SRAM) is = 0
					; Addressed via the Z-register + displacement
dbitz @0			
retz
.endmacro

.macro retz1		; Return if a bit in memory (SRAM) is = 1
					; Addressed via the Z-register + displacement
dbitz @0			
retnz
.endmacro


; **********************************************************************************************

; 					11.  Access memory and I/O with the Y-register as the base address register   **************

		
.macro litdy		; Store or output an immediate 8-bit value (literal)	
ldi  temp,@1		; displacement,literal
std y+@0,temp
.endm

.macro litwdy		; Store or output an immediate 16-bit value (literal)
ldi  temp,low(@1)	; displacement,literal
std y+@0,temp
ldi temp,high(@1)
std y+@0+1,temp
.endm

.macro lddy			; Load or input into a register (r0...r31)
ldd @0,y+@1			; register_adrs,displacement
.endmacro

.macro stdy			; Store or output a register content (r0...r31)
std y+@0,@1			; displacement,register_adrs
.endmacro

.macro ldady		; Load or input into the A-register (16 bits)
ldd a,y+@0			; displacement
ldd ah,y+@0+1
.endmacro

.macro ldbdy		; Load or input into the B-register (16 bits)
ldd b,y+@0			; displacement
ldd bh,y+@0+1
.endmacro

.macro ldcdy		; Load or input into the C-register (16 bits)
ldd a,y+@0			; displacement
ldd ah,y+@0+1
.endmacro

.macro ldxdy		; Load or input into the X-register (16 bits)
ldd xl,y+@0			; displacement
ldd xh,y+@0+1
.endmacro

.macro ldzdy		; Load or input into the Z-register (16 bits)
ldd zl,y+@0			; displacement
ldd zh,y+@0+1
.endmacro

.macro stady		; Store or output the content of the A-register (16 bits)
#if defined(lowbytefirst)
std y+@0,al		; displacement
std y+@0+1,ah
#else
std y+@0+1,ah		
std y+@0,al
#endif
.endmacro

.macro stbdy		; Store or output the content of the B-register (16 bits)
#if defined(lowbytefirst)
std y+@0,bl		; displacement
std y+@0+1,bh
#else
std y+@0+1,bh		
std y+@0,bl
#endif
.endmacro

.macro stcdy		; Store or output the content of the C-register (16 bits)
#if defined(lowbytefirst)
std y+@0,cl		; displacement
std y+@0+1,ch
#else
std y+@0+1,ch		
std y+@0,cl
#endif
.endmacro

.macro cleardy		; Clear the addressed byte		
eor temp,temp
std y+@0,temp		; displacement
.endm

.macro clearwdy		; Clear the addressed 16-bit word		
#if defined(lowbytefirst)
std y+@0,temp		; displacement
std y+@0+1,temp
#else
std y+@0+1,temp		
std y+@0,temp
#endif
.endm

.macro zerody		; Test, whether the addressed byte is zero		
ldd temp,y+@0		; displacement
and temp,temp
.endmacro

.macro zerowdy		; Test, whether the addressed 16-bit word is zero		
ldd temp,y+@0		; displacement
and temp,temp
brne e1				; if low-order byte not zero
ldd temp,y+@0+1		; high-order byte
and temp,temp
e1:
.endmacro

.macro bitdy0		; Clear the addressed bi
ldd temp,y+@0		; displacement,bit_adr		
andi temp, ~(1 << @1)
std y+@0,temp
.endmacro

.macro bitdy1	 	; Set the addressed bi		
ldd temp,y+@0		; displacement,bit_adr		
ori temp, 1 << @1
std y+@0,temp
.endmacro

.macro bitdy		; Test, whether the addressed bit is zero		
ldd temp,y+@0		; displacement,bit_adr	
andi temp, 1 << @1
.endmacro

.macro skipdy0  	; Skip the following instruction if the adressed bit = 0
ldd temp,y+@0		; displacement,bit_adr		
sbrc temp,@1
.endmacro

.macro skipdy1		; Skip the following instruction if the adressed bit = 1
ldd temp,y+@0		; displacement,bit_adr	
sbrs temp,@1
.endmacro


; **********************************************************************************************

; 					12.  Access bytes in SRAM by indirect addressing (all registers r0...r31 *************

;						 The address parameter is the address of the address pointer.	

.macro ldir			; Load indirect		
pushy				
lds yl,@1			; register_adrs, SRAM_adrs (address of the address pointer)
lds yh,@1+1
ld @0,y
popy
.endmacro

.macro ldirinc		; Load indirect. Post-increment address pointer
pushy
lds yl,@1			; register_adrs, SRAM_adrs (address of the address pointer)
lds yh,@1+1
ld @0,y+
sts @1,yl
sts @1+1,yh
popy
.endmacro

.macro ldirdec		; Pre-decrement address pointer. Then load indirect	
pushy
lds yl,@1			; register_adrs, SRAM_adrs (address of the address pointer)
lds yh,@1+1
ld @0,-y
sts @1,yl
sts @1+1,yh
popy
.endmacro

.macro ldirb		; Load indirect. Base address + displacement literal
pushy
lds yl,@1			; register_adrs, SRAM_adrs (address of the address pointer), displacement_literal
lds yh,@1+1
ldi temp,low(@2)
add yl,temp
ldi temp,high(@2)
adc yh,temp
ld @0,y
popy
.endmacro

.macro ldirbd		;  Load indirect. Base address + displacement address
pushy
lds yl,@1			; register_adrs, SRAM_adrs (address of the address pointer), displacement_adrs
lds yh,@1+1
lds temp,@2
add yl,temp
lds temp,@2+1
adc yh,temp
ld @0,y
popy
.endmacro

.macro stir			; Store indirect
pushy
lds yl,@0			; register_adrs, SRAM_adrs (address of the address pointer)
lds yh,@0+1
st y,@1
popy
.endmacro

.macro stirinc		; Store register indirect. Post-increment address pointer
pushy
lds yl,@0			; register_adrs, SRAM_adrs (address of the address pointer)
lds yh,@0+1
st y+,@1
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro stirdec		; Pre-decrement address pointer. Then store indirect	
pushy
lds yl,@0			; register_adrs, SRAM_adrs (address of the address pointer)
lds yh,@0+1
st -y,@1
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro stirb		; Store indirect. Base address + displacement literal
pushy
lds yl,@0			; register_adrs, SRAM_adrs (address of the address pointer), displacement_literal
lds yh,@0+1
ldi temp,low(@2)
add yl,temp
ldi temp,high(@2)
adc yh,temp
st @1,y
popy
.endmacro

.macro stirbd		; Store indirect. Base address + displacement address
pushy
push
lds yl,@0			; register_adrs, SRAM_adrs (address of the address pointer), displacement_adrs
lds yh,@0+1
lds temp,@2
add yl,temp
lds temp,@2+1
adc yh,temp
st @1,y
popy
.endmacro


; **********************************************************************************************

; 					13.  Access 16-bit words in  SRAM by indirect addressing ****************************

;						 The address parameter is the address of the address pointer.	

;			     	     Only valid for registers A, B, C, X, Y, Z

;						 Use only to access SRAM, not I/O registers, because the write order is not defined.


.macro ldiwa		; Load A-register indirect	
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
ld al,y+
ld ah,y
popy
.endmacro

.macro ldiwinca		; Load A-register indirect. Post-increment address pointer
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
ld al,y+
ld ah,y+
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro ldiwdeca		; Pre-decrement address pointer. Then load A-register indirect
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)	
lds yh,@0+1
ld ah,-y
ld al,-y
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro ldiwba		; Load A-register indirect. Base address + displacement literal
pushy
lds yl,@1			; register_adrs, SRAM_adrs (address of the address pointer), displacement_literal
lds yh,@1+1
ldi temp,low(@2)
add yl,temp
ldi temp,high(@2)
adc yh,temp
ld al,y+
ld ah,y
popy
.endmacro

.macro ldwbda		; Load A-register indirect. Base address + displacement address
pushy
lds yl,@1			; register_adrs, SRAM_adrs (address of the address pointer), displacement_adrs
lds yh,@1+1
lds temp,@2
add yl,temp
lds temp,@2+1
adc yh,temp
ld al,y+
ld ah,y
popy
.endmacro

.macro stiwa		; Store A-register indirect	
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
st y+,al
st y,ah
popy
.endmacro

.macro stiwinca		; Store A-register indirect. Post-increment address pointer
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
st y+,al
st y+,ah
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro stiwdeca		; Pre-decrement address pointer. Then store A-register indirect
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
st -y,ah
st -y,al
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro stiwba		; Store A-register indirect. Base address + displacement address
pushy
lds yl,@0			; register_adrs, SRAM_adrs (address of the address pointer), displacement_literal
lds yh,@0+1
ldi temp,low(@2)
add yl,temp
ldi temp,high(@2)
adc yh,temp
st al,y+
st ah,y
popy
.endmacro

.macro stiwbda		; Store A-register indirect. Base address + displacement address)
pushy
lds yl,@0			; register_adrs, SRAM_adrs (address of the address pointer), displacement_adrs
lds yh,@0+1
lds temp,@2
add yl,temp
lds temp,@2+1
adc yh,temp
st al,y+
st ah,y
popy
.endmacro

;

.macro ldiwb		; Load B-register indirect	
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
ld bl,y+
ld bh,y
popy
.endmacro

.macro ldiwincb		; Load B-register indirect. Post-increment address pointer
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
ld bl,y+
ld bh,y+
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro ldiwdecb		; Pre-decrement address pointer. Then load B-register indirect
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)	
lds yh,@0+1
ld bh,-y
ld bl,-y
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro ldiwbb		; Load B-register indirect. Base address + displacement literal
pushy
lds yl,@1			; register_adrs, SRAM_adrs (address of the address pointer), displacement_literal
lds yh,@1+1
ldi temp,low(@2)
add yl,temp
ldi temp,high(@2)
adc yh,temp
ld bl,y+
ld bh,y
popy
.endmacro

.macro ldwbdb		; Load B-register indirect. Base address + displacement address
pushy
lds yl,@1			; register_adrs, SRAM_adrs (address of the address pointer), displacement_adrs
lds yh,@1+1
lds temp,@2
add yl,temp
lds temp,@2+1
adc yh,temp
ld bl,y+
ld bh,y
popy
.endmacro

.macro stiwb		; Store B-register indirect	
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
st y+,bl
st y,bh
popy
.endmacro

.macro stiwincb		; Store B-register indirect. Post-increment address pointer
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
st y+,bl
st y+,bh
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro stiwdecb		; Pre-decrement address pointer. Then store A-register indirect
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
st -y,bh
st -y,bl
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro stiwbb		; Store B-register indirect. Base address + displacement address
pushy
lds yl,@0			; register_adrs, SRAM_adrs (address of the address pointer), displacement_literal
lds yh,@0+1
ldi temp,low(@2)
add yl,temp
ldi temp,high(@2)
adc yh,temp
st bl,y+
st bh,y
popy
.endmacro

.macro stiwbdb		; Store B-register indirect. Base address + displacement address)
pushy
lds yl,@0			; register_adrs, SRAM_adrs (address of the address pointer), displacement_adrs
lds yh,@0+1
lds temp,@2
add yl,temp
lds temp,@2+1
adc yh,temp
st bl,y+
st bh,y
popy
.endmacro

;

.macro ldiwc		; Load C-register indirect	
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
ld cl,y+
ld ch,y
popy
.endmacro

.macro ldiwincc		; Load C-register indirect. Post-increment address pointer
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
ld cl,y+
ld ch,y+
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro ldiwdecc		; Pre-decrement address pointer. Then load C-register indirect
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)	
lds yh,@0+1
ld ch,-y
ld cl,-y
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro ldiwbc		; Load C-register indirect. Base address + displacement literal
pushy
lds yl,@1			; register_adrs, SRAM_adrs (address of the address pointer), displacement_literal
lds yh,@1+1
ldi temp,low(@2)
add yl,temp
ldi temp,high(@2)
adc yh,temp
ld cl,y+
ld ch,y
popy
.endmacro

.macro ldwbdc		; Load C-register indirect. Base address + displacement address
pushy
lds yl,@1			; register_adrs, SRAM_adrs (address of the address pointer), displacement_adrs
lds yh,@1+1
lds temp,@2
add yl,temp
lds temp,@2+1
adc yh,temp
ld cl,y+
ld ch,y
popy
.endmacro

.macro stiwc		; Store A-register indirect	
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
st y+,cl
st y,ch
popy
.endmacro

.macro stiwincc		; Store C-register indirect. Post-increment address pointer
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
st y+,cl
st y+,ch
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro stiwdecc		; Pre-decrement address pointer. Then store C-register indirect
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
st -y,ch
st -y,cl
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro stiwbc		; Store C-register indirect. Base address + displacement address
pushy
lds yl,@0			; register_adrs, SRAM_adrs (address of the address pointer), displacement_literal
lds yh,@0+1
ldi temp,low(@2)
add yl,temp
ldi temp,high(@2)
adc yh,temp
st cl,y+
st ch,y
popy
.endmacro

.macro stiwbdc		; Store C-register indirect. Base address + displacement address)
pushy
lds yl,@0			; register_adrs, SRAM_adrs (address of the address pointer), displacement_adrs
lds yh,@0+1
lds temp,@2
add yl,temp
lds temp,@2+1
adc yh,temp
st cl,y+
st ch,y
popy
.endmacro

;

.macro ldiwx		; Load X-register indirect	
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
ld xl,y+
ld xh,y
popy
.endmacro

.macro ldiwincx		; Load X-register indirect. Post-increment address pointer
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
ld xl,y+
ld xh,y+
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro ldiwdecx		; Pre-decrement address pointer. Then load X-register indirect
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)	
lds yh,@0+1
ld xh,-y
ld xl,-y
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro ldiwbx		; Load X-register indirect. Base address + displacement literal
pushy
lds yl,@1			; register_adrs, SRAM_adrs (address of the address pointer), displacement_literal
lds yh,@1+1
ldi temp,low(@2)
add yl,temp
ldi temp,high(@2)
adc yh,temp
ld xl,y+
ld xh,y
popy
.endmacro

.macro ldwbdx		; Load X-register indirect. Base address + displacement address
pushy
lds yl,@1			; register_adrs, SRAM_adrs (address of the address pointer), displacement_adrs
lds yh,@1+1
lds temp,@2
add yl,temp
lds temp,@2+1
adc yh,temp
ld xl,y+
ld xh,y
popy
.endmacro

.macro stiwx		; Store X-register indirect	
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
st y+,xl
st y,xh
popy
.endmacro

.macro stiwincx		; Store X-register indirect. Post-increment address pointer
pushy
lds yl,@0				; SRAM_adrs (address of the address pointer)
lds yh,@0+1
st y+,xl
st y+,xh
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro stiwdecx		; Pre-decrement address pointer. Then store X-register indirect
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
st -y,xh
st -y,xl
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro stiwbx		; Store X-register indirect. Base address + displacement address
pushy
lds yl,@0			; register_adrs, SRAM_adrs (address of the address pointer), displacement_literal
lds yh,@0+1
ldi temp,low(@2)
add yl,temp
ldi temp,high(@2)
adc yh,temp
st xl,y+
st xh,y
popy
.endmacro

.macro stiwbdx		; Store X-register indirect. Base address + displacement address)
pushy
lds yl,@0			; register_adrs, SRAM_adrs (address of the address pointer), displacement_adrs
lds yh,@0+1
lds temp,@2
add yl,temp
lds temp,@2+1
adc yh,temp
st xl,y+
st xh,y
popy
.endmacro

;

.macro ldiwy		; Load Y-register indirect	
pushz
lds zl,@0			; SRAM_adrs (address of the address pointer)
lds zh,@0+1
ld yl,z+
ld yh,z
popz
.endmacro

.macro ldiwincy		; Load Y-register indirect. Post-increment address pointer
pushz
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
ld al,y+
ld ah,y+
sts @0,yl
sts @0+1,yh
popz
.endmacro

.macro ldiwdecy		; Pre-decrement address pointer. Then load Y-register indirect
pushz
lds zl,@0			; SRAM_adrs (address of the address pointer)	
lds zh,@0+1
ld yh,-z
ld yl,-z
sts @0,zl
sts @0+1,zh
popz
.endmacro

.macro ldiwby		; Load Y-register indirect. Base address + displacement literal
pushz
lds zl,@1			; register_adrs, SRAM_adrs (address of the address pointer), displacement_literal
lds zh,@1+1
ldi temp,low(@2)
add zl,temp
ldi temp,high(@2)
adc zh,temp
ld yl,z+
ld yh,z
popz
.endmacro

.macro ldwbdy		; Load Y-register indirect. Base address + displacement address
pushz
lds zl,@1			; register_adrs, SRAM_adrs (address of the address pointer), displacement_adrs
lds zh,@1+1
lds temp,@2
add zl,temp
lds temp,@2+1
adc zh,temp
ld yl,z+
ld yh,z
popz
.endmacro

.macro stiwy		; Store Y-register indirect	
pushz
lds zl,@0			; SRAM_adrs (address of the address pointer)
lds zh,@0+1
st z+,yl
st z,yh
popz
.endmacro

.macro stiwincy		; Store Y-register indirect. Post-increment address pointer
pushz
lds zl,@0			; SRAM_adrs (address of the address pointer)
lds zh,@0+1
st z+,yl
st z+,yh
sts @0,zl
sts @0+1,zh
popz
.endmacro

.macro stiwdecy		; Pre-decrement address pointer. Then store Y-register indirect
pushz
lds zl,@0			; SRAM_adrs (address of the address pointer)
lds zh,@0+1
st -z,yh
st -z,yl
sts @0,zl
sts @0+1,zh
popz
.endmacro

.macro stiwby		; Store Y-register indirect. Base address + displacement address
pushz
lds zl,@0			; register_adrs, SRAM_adrs (address of the address pointer), displacement_literal
lds zh,@0+1
ldi temp,low(@2)
add zl,temp
ldi temp,high(@2)
adc zh,temp
st yl,z+
st yh,z
popy
.endmacro

.macro stiwbdy		; Store Y-register indirect. Base address + displacement address)
pushy
lds zl,@0			; register_adrs, SRAM_adrs (address of the address pointer), displacement_adrs
lds zh,@0+1
lds temp,@2
add zl,temp
lds temp,@2+1
adc zh,temp
st yl,z+
st yh,z
popz
.endmacro

;

.macro ldiwz			; Load Z-register indirect	
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
ld zl,y+
ld zh,y
popy
.endmacro

.macro ldiwincz		; Load Z-register indirect. Post-increment address pointer
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
ld zl,y+
ld zh,y+
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro ldiwdecz		; Pre-decrement address pointer. Then load A-register indirect
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)	
lds yh,@0+1
ld zh,-y
ld zl,-y
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro ldiwbz		; Load Z-register indirect. Base address + displacement literal
pushy
lds yl,@1			; register_adrs, SRAM_adrs (address of the address pointer), displacement_literal
lds yh,@1+1
ldi temp,low(@2)
add yl,temp
ldi temp,high(@2)
adc yh,temp
ld zl,y+
ld zh,y
popy
.endmacro

.macro ldwbdz		; Load Z-register indirect. Base address + displacement address
pushy
lds yl,@1			; register_adrs, SRAM_adrs (address of the address pointer), displacement_adrs
lds yh,@1+1
lds temp,@2
add yl,temp
lds temp,@2+1
adc yh,temp
ld zl,y+
ld zh,y
popy
.endmacro

.macro stiwz		; Store Z-register indirect	
push yh
push yl
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
st y+,zl
st y,zh
pop yl
pop yh
.endmacro

.macro stiwincz		; Store Z-register indirect. Post-increment address pointer
pushy
lds yl,@0				; SRAM_adrs (address of the address pointer)
lds yh,@0+1
st y+,zl
st y+,zh
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro stiwdecz		; Pre-decrement address pointer. Then store Z-register indirect
pushy
lds yl,@0			; SRAM_adrs (address of the address pointer)
lds yh,@0+1
st -y,zh
st -y,zl
sts @0,yl
sts @0+1,yh
popy
.endmacro

.macro stiwbz		; Store Z-register indirect. Base address + displacement address
pushy
lds yl,@0			; register_adrs, SRAM_adrs (address of the address pointer), displacement_literal
lds yh,@0+1
ldi temp,low(@2)
add yl,temp
ldi temp,high(@2)
adc yh,temp
st zl,y+
st zh,y
popy
.endmacro

.macro stiwbdz		; Store Z-register indirect. Base address + displacement address)
pushy
lds yl,@0			; register_adrs, SRAM_adrs (address of the address pointer), displacement_adrs
lds yh,@0+1
lds temp,@2
add yl,temp
lds temp,@2+1
adc yh,temp
st zl,y+
st zh,y
popy
.endmacro

; **********************************************************************************************

; 					14.  Clear a 16-bit register (registers A, B, C, X, Y, Z) ****************************


.macro cleara		; Clear A-Register		
eor al,al
eor ah,ah
.endm

.macro clearb		; Clear B-Register	
eor r22,r22
eor r23,r23
.endm

.macro clearc		; Clear C-Register	
eor r20,r20
eor r21,r21
.endm

.macro clearx		; Clear X-Register	
eor r26,r26
eor r27,r27
.endm

.macro cleary		; Clear Y-Register	
eor r28,r28
eor r29,r29
.endm

.macro clearz		; Clear Z-Register	
eor r30,r30
eor r31,r31
.endm

; **********************************************************************************************

; 					15.  Load or store a 16.bit word (A, B, C, X, Y, Z) ***********

;					Macros with the general_mem_adrs parameter may be used to store into I/O registers,
;					because the write order is defined


.macro lda			; Load the 16-bit register A
gwld ah,al,@0		; general_mem_adrs
.endmacro

.macro ldbya		; Load the 16-bit register A with a zero-extended byte from the SRAM
lds ah,@0			; SRAM_adrs
ldi ah,0			; Zero extend		
.endmacro

.macro ldxa			; Load the 16-bit register A.  General memory address in X
ld al,x+
ld ah,x
sbiw xl,1
.endmacro

.macro ldya			; Load the 16-bit register A.  General memory address in Y
ld al,y
ldd ah,y+
.endmacro

.macro ldza			; Load the 16-bit register A.  General memory address in Z
ld al,z
ldd ah,z+
.endmacro

.macro sta			; Store the content of the 16-bit register A
gwst @0,ah,al
.endmacro

.macro stxa			; Store the content of the 16-bit register A. General memory address in X
#if defined(lowbytefirst)
st x+,al				
st x,ah
sbiw xl,1
#else
adiw xl,1
st x,ah		
st -x,al
#endif
.endmacro

.macro stya			; Store the content of the 16-bit register A. General memory address in Y
#if defined(lowbytefirst)
st y,al		
std y+,ah
sbiw xl,1
#else
adiw yl,1
std y,ah		
st -y,al
#endif
.endmacro

.macro stza			; Store the content of the 16-bit register A. General memory address in Z
#if defined(lowbytefirst)
st z,al		
std z+,ah
sbiw zl,1
#else
adiw zl,1
std z,ah		
st -z,al
#endif
.endmacro

.macro ldxinca		; Load the 16-bit register A. General memory address in X
					; Post-increment the address pointer in X by 2
ld al,x+
ld ah,x+
.endmacro

.macro ldyinca		; Load the 16-bit register A. General memory address in Y 
					; Post-increment the address pointer in Y by 2
ld al,y+
ld ah,y+
.endmacro

.macro stxinca		; Store the content of the 16-bit register A. General memory address in X
					; Post-increment the address pointer in X by 2
#if defined(lowbytefirst)
st x+,al				
st x+,ah
#else
adiw xl,1
st x,ah		
st -x,al
adiw xl,2
#endif
.endmacro

.macro styinca		; Store the content of the 16-bit register A. General memory address in Y
					; Post-increment the address pointer in Y by 2
#if defined(lowbytefirst)
st y+,al				
st y+,ah
#else
adiw xl,1
st y,ah		
st -y,al
adiw yl,2
#endif
.endmacro

;

.macro ldb			; Load the 16-bit register B
gdld bh,bl,@0		; general_mem_adrs
.endmacro

.macro ldbyb		; Load the 16-bit register B with a zero-extended byte from the SRAM
lds bh,@0			; SRAM_adrs
ldi bh,0			; Zero extend		
.endmacro

.macro ldxb			; Load the 16-bit register B.  General memory address in X
ld bl,x+
ld bh,x
sbiw xl,1
.endmacro

.macro ldyb			; Load the 16-bit register B.  General memory address in Y
ld bl,y
ldd bh,y+
.endmacro

.macro ldzb			; Load the 16-bit register B.  General memory address in Z
ld bl,z
ldd bh,z+
.endmacro

.macro stb			; Store the content of the 16-bit register B
gwst @0,bh,bl
.endmacro

.macro stxb			; Store the content of the 16-bit register B. General memory address in X
#if defined(lowbytefirst)
st x+,bl				
st x,bh
sbiw xl,1
#else
adiw xl,1
st x,bh		
st -x,bl
#endif
.endmacro

.macro styb			; Store the content of the 16-bit register B. General memory address in Y
#if defined(lowbytefirst)
st y,bl		
std y+,bh
sbiw xl,1
#else
adiw yl,1
std y,bh		
st -y,bl
#endif
.endmacro

.macro stzb			; Store the content of the 16-bit register B. General memory address in Z
#if defined(lowbytefirst)
st z,bl		
std z+,bh
sbiw zl,1
#else
adiw zl,1
std z,bh		
st -z,bl
#endif
.endmacro

.macro ldxincb		; Load the 16-bit register B. General memory address in X
					; Post-increment the address pointer in X by 2
ld bl,x+
ld bh,x+
.endmacro

.macro ldyincb		; Load the 16-bit register B. General memory address in Y 
					; Post-increment the address pointer in Y by 2
ld bl,y+
ld bh,y+
.endmacro

.macro stxincb		; Store the content of the 16-bit register B. General memory address in X
					; Post-increment the address pointer in X by 2
#if defined(lowbytefirst)
st x+,bl				
st x+,bh
#else
adiw xl,1
st x,bh		
st -x,bl
adiw xl,2
#endif
.endmacro

.macro styincb		; Store the content of the 16-bit register B. General memory address in Y
					; Post-increment the address pointer in Y by 2
#if defined(lowbytefirst)
st y+,bl				
st y+,bh
#else
adiw xl,1
st y,bh		
st -y,bl
adiw yl,2
#endif
.endmacro

;

.macro ldc			; Load the 16-bit register C
gdld ch,cl,@0		; general_mem_adrs
.endmacro

.macro ldbyc		; Load the 16-bit register C with a zero-extended byte from the SRAM
lds ch,@0			; SRAM_adrs
ldi ch,0			; Zero extend		
.endmacro

.macro ldxc			; Load the 16-bit register C.  General memory address in X
ld cl,x+
ld ch,x
sbiw xl,1
.endmacro

.macro ldyc			; Load the 16-bit register C.  General memory address in Y
ld cl,y
ldd ch,y+
.endmacro

.macro ldzc			; Load the 16-bit register C.  General memory address in Z
ld cl,y
ldd ch,y+
.endmacro

.macro stc			; Store the content of the 16-bit register C
gwst @0,ch,cl
.endmacro

.macro stxc			; Store the content of the 16-bit register C. General memory address in X
#if defined(lowbytefirst)
st x+,cl				
st x,ch
sbiw xl,1
#else
adiw xl,1
st x,ch		
st -x,cl
#endif
.endmacro

.macro styc			; Store the content of the 16-bit register C. General memory address in Y
#if defined(lowbytefirst)
st y,cl		
std y+,ch
sbiw xl,1
#else
adiw yl,1
std y,ch		
st -y,cl
#endif
.endmacro

.macro stzc			; Store the content of the 16-bit register C. General memory address in Z
#if defined(lowbytefirst)
st z,cl		
std z+,ch
sbiw zl,1
#else
adiw zl,1
std z,ch		
st -z,cl
#endif
.endmacro

.macro ldxincc		; Load the 16-bit register C. General memory address in X
					; Post-increment the address pointer in X by 2
ld cl,x+
ld ch,x+
.endmacro

.macro ldyincc		; Load the 16-bit register C. General memory address in Y 
					; Post-increment the address pointer in Y by 2
ld cl,y+
ld ch,y+
.endmacro

.macro stxincc		; Store the content of the 16-bit register CA. General memory address in X
					; Post-increment the address pointer in X by 2
#if defined(lowbytefirst)
st x+,cl				
st x+,ch
#else
adiw xl,1
st x,ch		
st -x,cl
adiw xl,2
#endif
.endmacro

.macro styincc		; Store the content of the 16-bit register C. General memory address in Y
					; Post-increment the address pointer in Y by 2
#if defined(lowbytefirst)
st y+,cl				
st y+,ch
#else
adiw xl,1
st y,ch		
st -y,cl
adiw yl,2
#endif
.endmacro

;

.macro ldx			; Load the 16-bit register X
gdld xh,xl,@0		; general_mem_adrs
.endmacro

.macro ldyx			; Load the 16-bit register X.  General memory address in Y
ld xl,y
ldd xh,y+
.endmacro

.macro ldzx			; Load the 16-bit register X.  General memory address in Z
ld xl,z
ldd xh,z+
.endmacro

.macro stx			; Store the content of the 16-bit register X
gwst @0,xh,xl
.endmacro

.macro styx			; Store the content of the 16-bit register X. General memory address in Y
#if defined(lowbytefirst)
st y,xl		
std y+,xh
sbiw yl,1
#else
adiw yl,1
std y,xh		
st -y,xl
#endif
.endmacro

.macro stzx			; Store the content of the 16-bit register X. General memory address in Z
#if defined(lowbytefirst)
st z,xl		
std z+,xh
sbiw zl,1
#else
adiw zl,1
std z,xh
st -z,xl
#endif
.endmacro

;

.macro ldy			; Load the 16-bit register Y
gdld yh,yl,@0		; general_mem_adrs
.endmacro

.macro ldxy			; Load the 16-bit register Y.  General memory address in X
ld yl,x
ldd yh,x+
.endmacro

.macro ldzy			; Load the 16-bit register Y.  General memory address in Z
ld yl,z
ldd yh,z+
.endmacro

.macro sty			; Store the content of the 16-bit register Y
gwst @0,yh,yl
.endmacro

.macro stxy			; Store the content of the 16-bit register Y. General memory address in X
#if defined(lowbytefirst)
st x,yl		
std x+,yh
sbiw xl,1
#else
adiw xl,1
std x,yh		
st -x,yl
#endif
.endmacro

.macro stzy			; Store the content of the 16-bit register Y. General memory address in Z
#if defined(lowbytefirst)
st z,yl		
std z+,yh
sbiw zl,1
#else
adiw zl,1
std z,yh
st -z,yl
#endif
.endmacro

;

.macro ldz			; Load the 16-bit register Z
gdld zh,zl,@0		; general_mem_adrs
.endmacro

.macro ldxz			; Load the 16-bit register Z.  General memory address in X
ld zl,x
ldd zh,x+
.endmacro

.macro ldyz			; Load the 16-bit register Z.  General memory address in Y
ld zl,y
ldd zh,y+
.endmacro

.macro stz			; Store the content of the 16-bit register Z
gwst @0,zh,zl
.endmacro

.macro stxz			; Store the content of the 16-bit register Z. General memory address in X
#if defined(lowbytefirst)
st x,zyl		
std x+,zh
sbiw xl,1
#else
adiw xl,1
std x,zh		
st -x,zl
#endif
.endmacro

.macro styz			; Store the content of the 16-bit register Z. General memory address in Y
#if defined(lowbytefirst)
st y,yz		
std y+,zh
sbiw yl,1
#else
adiw yl,1
std y,zh
st -y,zl
#endif
.endmacro


; **********************************************************************************************

; 					16.  Operations with 16-bit literals (registers A, B, C, X, Y, Z) ***********

.macro lita			; Load literal into A-register
ldi al,low(@0)		; 16-bit-literal
ldi ah,high(@0)
.endmacro

.macro addlita		; Add literal to A-register
ldi temp,low(@0)	; 16-bit-literal
add al,temp
ldi temp,high(@0)
adc ah,temp
.endmacro 

.macro sublita		; Subtract literal from A-register (A - Literal)
ldi temp,low(@0)	; 16-bit-literal
sub al,temp
ldi temp,high(@0)
sbc ah,temp
.endmacro 

.macro cmplita		; Compare literal with A-register (A - Literal)
ldi temp,low(@0)	; 16-bit-literal
cp al,temp
ldi temp,high(@0)
cpc ah,temp
.endmacro 

.macro andlita		; Bitwise AND literal with A-register
andi al,low(@0)		; 16-bit-literal
andi ah,high(@0)
.endmacro 

.macro orlita		; Bitwise OR literal with A-register
ori al,low(@0)		; 16-bit-literal
ori ah,high(@0)
.endmacro 

.macro xorlita		; Bitwise XOR literal with A-register
ldi temp,low(@0)	; 16-bit-literal
eor al,temp
ldi temp,high(@0)
eor ah,temp
.endmacro 

.macro zeroa		; Test A-register for zero content (ZF)		
and al,al
brne e1
and ah,ah
e1:
.endmacro

;

.macro litb			; Load literal into B-register
ldi bl,low(@0)		; 16-bit-literal
ldi bh,high(@0)
.endmacro

.macro addlitb		; Add literal to B-register
ldi temp,low(@0)	; 16-bit-literal
add r22,temp
ldi temp,high(@0)
adc r23,temp
.endmacro 

.macro sublitb		; Subtract literal from B-register (B - Literal)
ldi temp,low(@0)	; 16-bit-literal
sub r22,temp
ldi temp,high(@0)
sbc r23,temp
.endmacro 

.macro cmplitb		; Compare literal with B-register (B - Literal)
ldi temp,low(@0)	; 16-bit-literal
cp r22,temp
ldi temp,high(@0)
cpc r23,temp
.endmacro 

.macro andlitb		; Bitwise AND literal with B-register
andi r22,low(@0)	; 16-bit-literal
andi r23,high(@0)
.endmacro 

.macro orlitb		; Bitwise OR literal with B-register
ori r22,low(@0)		; 16-bit-literal
ori r23,high(@0)
.endmacro 

.macro xorlitb		; Bitwise XOR literal with B-register
ldi temp,low(@0)	; 16-bit-literal
eor r22,temp
ldi temp,high(@0)
eor r23,temp
.endmacro

.macro zerob		; Test B-register for zero content (ZF)	
and r22,r22
brne e1
and r23,r23
e1:
.endmacro

;

.macro litc			; Load literal into C-register
ldi cl,low(@0)		; 16-bit-literal
ldi ch,high(@0)
.endmacro

.macro addlitc		; Add literal to C-register
ldi temp,low(@0)	; 16-bit-literal
add r20,temp
ldi temp,high(@0)
adc r21,temp
.endmacro 

.macro sublitc		; Subtract literal from C-register (C - Literal)
ldi temp,low(@0)	; 16-bit-literal
sub r20,temp
ldi temp,high(@0)
sbc r21,temp
.endmacro 

.macro cmplitc		; Compare literal to C-register (C - Literal)
ldi temp,low(@0)	; 16-bit-literal
cp r20,temp
ldi temp,high(@0)
cpc r21,temp
.endmacro 

.macro andlitc		; Bitwise AND literal with C-register
andi r20,low(@0)	; 16-bit-literal
andi r21,high(@0)
.endmacro 

.macro orlitc		; Bitwise OR literal with B-register
ori r20,low(@0)		; 16-bit-literal
ori r21,high(@0)
.endmacro 

.macro xorlitc		; Bitwise XOR literal with B-register
ldi temp,low(@0)	; 16-bit-literal
eor r20,temp
ldi temp,high(@0)
eor r21,temp
.endmacro

.macro zeroc		; Test C-register for zero content (ZF)	
and r20,r20
brne e1
and r21,r21
e1:
.endmacro

; 

.macro litx			; Load literal into X-register	
ldi xl,low(@0)		; 16-bit-literal
ldi xh,high(@0)
.endmacro

.macro addlitx		; Add literal to X-register
ldi temp,low(@0)	; 16-bit-literal
add r26,temp
ldi temp,high(@0)
adc r27,temp
.endmacro 

.macro sublitx		; Subtract literal from X-register (X - Literal)
ldi temp,low(@0)	; 16-bit-literal
sub r26,temp
ldi temp,high(@0)
sbc r27,temp
.endmacro 

.macro cmplitx		; Compare literal with X-register (X - Literal)
ldi temp,low(@0)	; 16-bit-literal
cp r26,temp
ldi temp,high(@0)
cpc r27,temp
.endmacro 

.macro zerox		; Test X-register for zero content (ZF)		
and r26,r26
brne e1
and r27,r27
e1:
.endmacro

;

.macro lity			; Load literal into Y-register
ldi yl,low(@0)		; 16-bit-literal
ldi yh,high(@0)
.endmacro

.macro addlity		; Add literal to Y-register
ldi temp,low(@0)	; 16-bit-literal
add r28,temp
ldi temp,high(@0)
adc r29,temp
.endmacro 

.macro sublity		; Subtract literal from Y-register (Y - Literal)
ldi temp,low(@0)	; 16-bit-literal
sub r28,temp
ldi temp,high(@0)
sbc r29,temp
.endmacro 

.macro cmplity		; Compare literal with Y-register (Y - Literal)
ldi temp,low(@0)	; 16-bit-literal
cp r28,temp
ldi temp,high(@0)
cpc r29,temp
.endmacro
 
.macro zeroy		; Test X-register for zero content (ZF)		
and r28,r28
brne e1
and r29,r29
e1:
.endmacro

;

.macro litz			; Load literal into Z-register
ldi zl,low(@0)		; 16-bit-literal
ldi zh,high(@0)
.endmacro

.macro addlitz		; Add literal to Z-register
ldi temp,low(@0)	; 16-bit-literal
add r30,temp
ldi temp,high(@0)
adc r31,temp
.endmacro 

.macro sublitz		;  Subtract literal from Z-register (Z - Literal)
ldi temp,low(@0)	; 16-bit-literal
sub r30,temp
ldi temp,high(@0)
sbc r31,temp
.endmacro 

.macro cmplitz		;  Compare literal with Z-register (Z - Literal)
ldi temp,low(@0)	; 16-bit-literal
cp r30,temp
ldi temp,high(@0)
cpc r31,temp
.endmacro
 
.macro zeroz		; Test Z-register for zero content (ZF)	
and r30,r30
brne e1
and r31,r31
e1:
.endmacro

; **********************************************************************************************

; 					17.  Operations with 8-bit literals and registers r0 to r31 ***********


.macro addlit		; Add literal to register	
ldi temp,@1			; register address, literal
add @0,temp
.endmacro

.macro adclit		; Add literal + Carry-In (CF) to register 	
ldi temp,@1			; register address, literal
adc @0,temp
.endmacro

.macro sublit		; Subtract literal from register (register - literal)	
ldi temp,@1			; register address, literal
sub @0,temp
.endmacro

.macro sbclit		; Subtract literal - Borrow-In (CF) from register (register - literal - CF)	
ldi temp,@1			; register address, literal
sbc @0,temp
.endmacro

.macro cmplit 		; Compare literal with register (register - literal)			
ldi temp,@1			; register address, literal
cp @0,temp
.endmacro

.macro cmpclit		; Compare literal - Borrow-In (CF) with register (register - literal - CF)		
ldi temp,@1			; register address, literal
cpc @0,temp
.endmacro

.macro andlit		; Bitwise AND literal with register	
ldi temp,@1			; register address, literal
and @0,temp
.endmacro

.macro orlit		; Bitwise OR literal with register			
ldi temp,@1			; register address, literal
or @0,temp
.endmacro

.macro xorlit		; Bitwise XOR literal with register		
ldi temp,@1			; register address, literal
eor @0,temp
.endmacro


; **********************************************************************************************

; 					18.  Operations between 16-bit registers ***********************************


.macro addab		; Add B-register to A-register. A = A + B		
add al,r22
adc ah,r23
.endmacro

.macro subab		; Subtract B-register from A-register. A = A - B				
sub al,r22
sbc ah,r23
.endmacro

.macro cmpab		; Compare  B-register with A-register. A - B	
cp al,r22
cpc ah,r23
.endmacro

.macro andab		; Bitwise AND B-register and A-register. A = A AND B
and al,r22
and ah,r23
.endmacro

.macro orab			; Bitwise OR B-register and A-register. A = A OR B
or al,r22
or ah,r23
.endmacro

.macro xorab		; Bitwise XOR B-register and A-register. A = A XOR B		
eor al,r22
eor ah,r23
.endmacro

.macro addac		; Add C-register to A-register. A = A + C		
add al,r20
adc ah,r21
.endmacro

.macro subac		; Subtract C-register from A-register. A = A - C		
sub al,r20
sbc ah,r21
.endmacro

.macro cmpac		; Compare  C-register with A-register. A - C			
cp al,r20
cpc ah,r21
.endmacro

.macro andac		; Bitwise AND C-register and A-register. A = A AND C	
and al,r20
and ah,r21
.endmacro

.macro orac			; Bitwise OR C-register and A-register. A = A OR C	
or al,r20
or ah,r21
.endmacro

.macro xorac		; Bitwise XOR C-register and A-register. A = A XOR C		
eor al,r20
eor ah,r21
.endmacro

.macro addax		; Add X-register to A-register. A = A + X
add al,xl
adc ah,xh
.endmacro

.macro subax		; Subtract X-register from A-register. A = A - X	
sub al,xl
sbc ah,xh
.endmacro

.macro cmpax		; Compare  X-register with A-register. A - X	
cp al,xl
cpc ah,xh
.endmacro

.macro adday		; Add Y-register to A-register. A = A + Y	
add al,yl
adc ah,yh
.endmacro

.macro subay		; Subtract Y-register from A-register. A = A - Y	
sub al,yl
sbc ah,yh
.endmacro

.macro cmpay		; Compare  Y-register with A-register. A - Y	
cp al,yl
cpc ah,yh
.endmacro

.macro addaz		; Add Z-register to A-register. A = A + Z	
add al,zl
adc ah,zh
.endmacro

.macro subaz		; Subtract Z-register from A-register. A = A - Z	
sub al,zl
sbc ah,zh
.endmacro

.macro cmpaz		; Compare  Z-register with A-register. A - Z	
cp al,zl
cpc ah,zh
.endmacro

.macro addxa		; Add A-register to X-register. X = X + A (address calculation)	
add xl,al
adc xh,ah
.endmacro

.macro subxa		; Subtract A-register from X-register. X = X - A (address calculation)		
sub xl,al
sbc xh,ah
.endmacro

.macro cmpxa		; Compare A-register with X-register. X = X - A (address calculation)		
cp xl,al
cpc xh,ah
.endmacro

.macro addya		; Add A-register to Y-register. Y = Y + A (address calculation)		
add yl,al
adc yh,ah
.endmacro

.macro subya		; Subtract A-register from Y-register. Y = Y - A (address calculation)		
sub yl,al
sbc yh,ah
.endmacro

.macro cmpya		; Compare A-register with Y-register. Y = Y - A (address calculation)		
cp yl,al
cpc yh,ah
.endmacro

.macro addza		; Add A-register to Z-register. Z = Z + A (address calculation)			
add zl,al
adc zh,ah
.endmacro

.macro subza		; Subtract A-register from Z-register. Z = Z - A (address calculation)		
sub zl,al
sbc zh,ah
.endmacro

.macro cmpza		; Compare A-register with Z-register. Z = Z - A (address calculation)	
cp zl,al
cpc zh,ah
.endmacro

.macro lsla			; Shift A-register one bit to the left. Fill with 0	
lsl al
rol ah
.endmacro

.macro lsla1		; Shift A-register one bit to the left. Fill with 1			
lsl al
rol ah
ori al,1
.endmacro

.macro lsra			; Shift A-register one bit to the right. Fill with 0	
lsr ah
ror al
.endmacro

.macro lsra1		; Shift A-register one bit to the right. Fill with 1		
lsr ah
ror al
ori ah,0x80
.endmacro

.macro asra			; Shift A-register arithmetically one bit to the right. Retain the sign bit
asr ah
ror al
.endmacro

.macro rola			; Rotate A-register one bit to the left. 16-bit rotation; from bit 15 into bit 0		
lsl al
rol ah
brcc e1
ori al,1
e1:
.endmacro

.macro rora			; Rotate A-register one bit to the right. 16-bit rotation; from bit 0 into bit 15
lsr ah
ror al
brcc e1
ori ah,0x80
e1:
.endmacro

.macro nega			; Change sign in A-register (Two's complement)	
com al
com ah
adiw al,1
.endmacro

.macro absa			; Absolute value in A-register
sbrs ah,7			; test sign flag
rjmp e1				; not set; positive
nega				; set. Negative. Change sign
e1:
.endmacro


; **********************************************************************************************

; 					19.  Operations between 16-bit and 8-bit registers *************************

;						  The content of the 8-bit register will be zero-extended to 16 bits



.macro addaby		; Add 8-bit-register to A-register. A = A + r
eor temp,temp
add al,@0			; register_adrs
adc ah,temp
.endmacro

.macro subaby		; Subtract 8-bit-register from A-register. A = A - r
eor temp,temp
sub al,@0			; register_adrs
sbc ah,temp
.endmacro

.macro cmpbaby		; Compare 8-bit-register with A-register. A - r
eor temp,temp
cp al,@0			; register_adrs
cpi ah,temp
.endmacro

.macro addbby		; Add 8-bit-register to B-register. B = B + r
eor temp,temp
add al,@0			; register_adrs
adc ah,temp
.endmacro

.macro subbby		; Subtract 8-bit-register from B-register. B = B - r
eor temp,temp
sub al,@0			; register_adrs
sbc ah,temp
.endmacro

.macro cmpbbby		; Compare 8-bit-register with B-register. B - r
eor temp,temp
cp bl,@0			; register_adrs
cpi bh,temp
.endmacro

.macro addcby		; Add 8-bit-register to C-register. C = C + r
eor temp,temp
add al,@0			; register_adrs
adc ah,temp
.endmacro

.macro subcby		; Subtract 8-bit-register from C-register. C = C - r
eor temp,temp
sub al,@0			; register_adrs
sbc ah,temp
.endmacro

.macro cmpbcby		; Compare 8-bit-register with C-register. C - r
eor temp,temp
cp cl,@0			; register_adrs
cpi ch,temp
.endmacro

.macro addxby		; Add 8-bit-register to X-register. X = X + r
eor temp,temp
add xl,@0			; register_adrs
adc xh,temp
.endmacro

.macro subxby		; Subtract 8-bit-register from X-register. X= X - r
eor temp,temp
sub xl,@0			; register_adrs
sbc xh,temp
.endmacro

.macro cmpxby		; Compare 8-bit-register with XA-register. X - r
eor temp,temp
cp xl,@0			; register_adrs
cpi xh,temp
.endmacro

.macro addyby		; Add 8-bit-register to Y-register. Y = Y + r
eor temp,temp
add yl,@0			; register_adrs
adc yh,temp
.endmacro

.macro subyby		; Subtract 8-bit-register from Y-register. Y = Y - r
eor temp,temp
sub yl,@0			; register_adrs
sbc yh,temp
.endmacro

.macro cmpbyby		; Compare 8-bit-register with Y-register. Y - r
eor temp,temp
cp yl,@0			; register_adrs
cpi yh,temp
.endmacro

.macro addzby		; Add 8-bit-register to Z-register. Z = Z + r
eor temp,temp
add zl,@0; register_adrs
adc zh,temp
.endmacro

.macro subzby		; Subtract 8-bit-register from Z-register. Z = Z - r
eor temp,temp
sub zl,@0			; register_adrs
sbc zh,temp
.endmacro

.macro cmpbzby		; Compare 8-bit-register with Z-register. Z - r
eor temp,temp
cp zl,@0			; register_adrs
cpi zh,temp
.endmacro


; **********************************************************************************************

; 					20.  Operations 8-bit register - memory (SRAM) ***********


.macro adds			; ADD byte from SRAM_adrs to register. r = r + <SRAM>
lds temp,@1			; register_adrs, SRAM_adrs
add @0,temp
.endmacro

.macro subs			; Subtract byte from SRAM_adrs from register. r = r - <SRAM>
lds temp,@1			; register_adrs, SRAM_adrs
sub @0,temp
.endmacro

.macro cmps			; Compare byte from SRAM_adrs with register. r - <SRAM>
lds temp,@1			; register_adrs, SRAM_adrs
cp @0,temp
.endmacro

.macro addsx		; ADD SRAM byte from X+displacement_adrs to register. r = r + <SRAM
ld temp,x+@1		; register_adrs, displacement_adrs
add @0,temp
.endmacro

.macro subsx		; Subtract SRAM byte from X+displacement_adrs from register. r = r - <SRAM>
lds temp,x+@1		; register_adrs, displacement_adrs
sub @0,temp
.endmacro

.macro cmpsx 		; Compare SRAM byte from X+displacement_adrs with register. r - <SRAM>
lds temp,x+@1		; register_adrs, displacement_adrs
cp @0,temp
.endmacro

.macro addsy		; ADD SRAM byte from Y+displacement_adrs to register. r = r + <SRAM
ld temp,y+@1		; register_adrs, displacement_adrs
add @0,temp
.endmacro

.macro subsy 		; Subtract SRAM byte from Y+displacement_adrs from register. r = r - <SRAM>
lds temp,y+@1		; register_adrs, displacement_adrs
sub @0,temp
.endmacro

.macro cmpsy		; Compare SRAM byte from Y+displacement_adrs witj register. r - <SRAM>
lds temp,y+@1		; register_adrs, displacement_adrs
cp @0,temp
.endmacro


; **********************************************************************************************

; 					21.  Operations 16-bit register - memory (SRAM) ***********

; 							Byte operands will be zero-extended to 16 bits


.macro addbyas		; Add SRAM byte to A-register (16 Bits). A = A + <SRAM>
lds temp,@0			; SRAM_adrs
add a,temp
ldi temp,0
adc ah,temp
.endmacro

.macro subbyas		; Subtract SRAM byte from A-register (16 Bits). A = A - <SRAM>	
lds temp,@0			; SRAM_adrs
sub a,temp
ldi temp,0
sbc ah,temp
.endmacro

.macro cmpbyas		; Compare SRAM byte with A-register (16 Bits). A - <SRAM>				
lds temp,@0			; SRAM_adrs
cp a,temp
ldi temp,0
cpc ah,temp
.endmacro

.macro addbybs		; Add SRAM byte to B-register (16 Bits). B = B + <SRAM>
lds temp,@0			; SRAM_adrs
add b,temp
ldi temp,0
adc bh,temp
.endmacro

.macro subbybs		; Subtract SRAM byte from B-register (16 Bits). B = B - <SRAM>
lds temp,@0			; SRAM_adrs
sub b,temp
ldi temp,0
sbc bh,temp
.endmacro

.macro cmpbybs		; Compare SRAM byte with B-register (16 Bits). B - <SRAM>	
lds temp,@0			; SRAM_adrs
cp b,temp
ldi temp,0
cpc bh,temp
.endmacro

.macro addwas		; Add SRAM word to A-register (16 Bits). A = A + <SRAM>	
lds temp,@0			; SRAM_adrs
add a,temp
lds temp,@0+1
adc ah,temp
.endmacro

.macro subwas		; Subtract SRAM word from A-register (16 Bits). A = A - <SRAM>
lds temp,@0			; SRAM_adrs
sub a,temp
lds temp,@0+1
sbc ah,temp
.endmacro

.macro cmpwas		; Compare SRAM word with A-register (16 Bits). A - <SRAM>	
lds temp,@0			; SRAM_adrs
cp a,temp
lds temp,@0+1
cpc ah,temp
.endmacro

.macro addwbs		; Add SRAM word to B-register (16 Bits). B = B + <SRAM>		
lds temp,@0			; SRAM_adrs
add b,temp
lds temp,@0+1
adc bh,temp
.endmacro

.macro subwbs		; Subtract SRAM word from B-register (16 Bits). B = B - <SRAM>
lds temp,@0			; SRAM_adrs
sub b,temp
lds temp,@0+1
sbc bh,temp
.endmacro

.macro cmpwbs		; Compare SRAM word with B-register (16 Bits). B - <SRAM>	
lds temp,@0			; SRAM_adrs
cp b,temp
lds temp,@0+1
cpc bh,temp
.endmacro


; **********************************************************************************************

; 					22.  Operations with 24- and 32-bit operands ***********


.macro lit24			; Load a 24-bit literal into the SRAM
ldi temp,low(@1)		; SRAM_adrs, 24-bit-literal
sts @0,temp
ldi temp,byte2(@1)
sts @0+1,temp
ldi temp,byte3(@1)
sts @0+2,temp
.endmacro

.macro lit32			; Load a 32-bit literal into SRAM
ldi temp,low(@1)		; SRAM_adrs, 32-bit-literal
sts @0,temp
ldi temp,byte2(@1)
sts @0+1,temp
ldi temp,byte3(@1)
sts @0+2,temp
ldi temp,byte4(@1)
sts @0+3,temp
.endmacro 

.macro litab32			; Load  a 32-bit literal into the registers A  and B
ldi al,low(@0)			; 32-bit-literal
ldi ah,byte2(@0)		; low-order word into A, high-order word into B
ldi bl,byte3(@0)
ldi bh,byte4(@0)
.endmacro 

.macro dec24			; Decrement a 3-byte word in SRAM by 1
push a
ldi a,1
lds temp,@0				; SRAM_adrs
sub temp,a
sts @0,temp
lds temp,@0+1
ldi a,0
sbc temp,a
sts @0+1,temp
lds temp,@0+2
sbc temp,a
sts @0+2,temp
pop a
.endmacro

.macro litc3			; Load a 24-bit-literal into the extended C-register
ldi cl,low(@0)			; 24-bit-literal
ldi ch,high(@0)
ldi temp,byte3(@0)
mov cx3,temp
.endmacro

.macro addlitc3			; Add a 24-bit-literal to the extended C-register. Cx = Cx + literal
ldi temp,low(@0)		; 24-bit-literal
add cl,temp
ldi temp,high(@0)
adc ch,temp
ldi temp,byte3(@0)
adc cx3,temp
.endmacro 

.macro sublitc3			; Subtract a 24-bit-literal from the extended C-register. Cx = Cx - literal
ldi temp,low(@0)		; 24-bit-literal
cp cl,temp
ldi temp,high(@0)
cpc ch,temp
ldi temp,byte3(@0)
cpc cx3,temp
.endmacro 

.macro cmplitc3			; Compare a 24-bit-literal with the extended C-register. Cx - literal
ldi temp,low(@0)
cp cl,temp
ldi temp,high(@0)
cpc ch,temp
ldi temp,byte3(@0)
cpc cx3,temp
.endmacro 

.macro inccx			; Increment the content of the extended C-register by 1
ldi temp,1		
add cl,temp
ldi temp,0
adc ch,temp
adc cx3,temp
.endmacro 

.macro deccx			; Decrement the content of the extended C-register by 1
ldi temp,1		
sub cl,temp
ldi temp,0
sbc ch,temp
sbc cx3,temp
.endmacro 


; **********************************************************************************************

; 					23.  Load a register from Flash or SRAM (load from a unified address) ***********

;						Address pointer in Z-register. Auto-incremnet after memory access
;						Bit 15 = 0: Flash
;						Bit 15 = 1: SRAM 	


.macro lduz 		; Load register from Flash or SRAM
					; SRAM_adrs
sbrc r31,7			; adrs bit 15 in Z?
rjmp lduz_sram		; High. Load from SRAM
lpm @0,z+			; Load from flash 
rjmp e1
lduz_sram:			
bitf0,r31			; clear bit 15 of adrs in Z
ld @0,z+			; load from SRAM
e1:
.endmacro

.macro lduzo		; Load register from Flash or SRAM. Address augmented by a 16-bit displacement
					; SRAM_adrs, 16-bit-displacement
pushz				; save Z
addlitz @1			; add the displacement
lduz @1				; load register
popz				; restore Z
.endmacro


; **********************************************************************************************

; 					24.  Delays ***********


.macro microsecs		; Microseoonds according to literal value (16 bits)
pusha
lita @0					; 16-bit-literal
cal mikros				; a device-specific subroutine
popa
.endmacro

.macro millisecs		; Milliseconds according to literal value (16 bits)
pusha
lita @0					; 16-bit-literal
cal millis				; a device-specific subroutine
popa
.endmacro

.macro microseca		; Microseoonds according to value in A-register (16 bits)
pusha
cal mikros
popa
.endmacro

.macro milliseca		; Milliseconds according to value in A-register (16 bits)
pusha
cal millis
popa
.endmacro


; **********************************************************************************************

; 						25.  Specialties  ******************************************************


.macro absreg			; Absolute amount of register content
sbrc @0,7				; Check sign bit
neg @0					; Set. Negative. Negate
.endmacro

.macro memlitz			; Store a 8-bit literal on an address provided in Z-register and increment address by 1
ldi temp,low(@0)		; literal
st z+,temp
.endmacro

.macro memwlitz			; Store a 16-bit literal on an address provided in Z-register and increment address by 2
ldi temp,high(@0)		; 16_bit_literal
st z+,temp
ldi temp,high(@0)
st z+,temp
.endmacro

.macro incmem			; Increment a byte stored in SRAM by 1
lds temp,@0				; SRAM_adrs
inc temp
sts @0,temp
.endmacro

.macro incwmem			; Increment a word stored in SRAM by 1	 
pusha
lds al,@0				; SRAM_adrs 
lds ah,@0+1
adiw al,1
sts @0,al
sts @0+1,ah
popa
.endmacro

.macro decmem			; Decrement a byte stored in SRAM by 1		
lds temp,@0				; SRAM_adrs
dec temp
sts @0,temp
.endmacro

.macro decwmem			; Decrement a word stored in SRAM by 1 
pusha
lds al,@0				; SRAM_adrs 
lds ah,@0+1
sbiw al,1
sts @0,al
sts @0+1,ah
popa
.endmacro

.macro addmemlit		; Add a literal to a byte stored in SRAM. <SRAM> = <SRAM> + literal
lds temp,@0				; SRAM_adrs,literal
subi temp,-@1
sts @0,temp
.endmacro

.macro submemlit		; Subtract a literal from a byte stored in SRAM.  <SRAM> = <SRAM> - literal
lds temp,@0				; SRAM_adrs,literal
subi temp,@1
sts @0,temp
.endmacro

.macro cmpmemlit		; Compare a literal with a byte stored in SRAM.  <SRAM> - literal
lds temp,@0				; SRAM_adrs,literal
cpi temp,@1
.endmacro

.macro addwmemlit		; Add a literal to a word stored in SRAM. <SRAM> = <SRAM> + literal
pusha
lds temp,@0				; SRAM_adrs,16_bit_literal 
ldi a,low(@1)
add temp,a
sts @0,temp
lds temp,@0+1
ldi a,high(@1)
adc temp,a
sts @0+1,temp
popa
.endmacro

.macro subwmemlit		; Subtract a literal from a word stored in SRAM.  <SRAM> = <SRAM> - literal  
pusha
lds temp,@0				; SRAM_adrs,16_bit_literal
ldi a,low(@1)
sub temp,a
sts @0,temp
lds temp,@0+1
ldi a,high(@1)
sbc temp,a
sts @0+1,temp
popa
.endmacro

.macro cmpwmemlit		; Compare a literal with a word stored in SRAM.  <SRAM> - literal  
pusha
lds temp,@0				; SRAM_adrs,16_bit_literal
ldi a,low(@1)
sub temp,a
lds temp,@0+1
ldi a,high(@1)
sbc temp,a
popa
.endmacro

.macro cmpa			; Compare the content of AL-register (8 bits) with  a byte stored in SRAM.  <SRAM> - AL
lds temp,@0			; SRAM_adrs
cp  al,temp
.endmacro

.macro cmpwa		; Compare the content of A-register (16 bits) with  a word stored in SRAM.  <SRAM> - A
lds temp,@0;		; SRAM_adrs
cp  al,temp
lds temp,@0+1
cpc ah,temp
.endmacro

.macro pointx			; Load a pointer into X-register. X = 1st SRAM_adrs + <2nd SRAM_adrs>
lds xl,@1				; The first parameter is the starting address, e.g., of an array
ldi xh,0				; The 2nd address points to an SRAM location containing the displacement value
ldi temp,low(@0)		; SRAM_adrs, SRAM_adrs
add xl,temp
ldi temp,high(@0)
adc xh,temp
.endmacro

.macro pointy			; Load a pointer into Y-register. Y = 1st SRAM_adrs + <2nd SRAM_adrs>
lds yl,@1				; The first parameter is the starting address, e.g., of an array
ldi yh,0				; The 2nd address points to an SRAM location containing the displacement value
ldi temp,low(@0)		; SRAM_adrs, SRAM_adrs
add yl,temp
ldi temp,high(@0)
adc yh,temp
.endmacro

.macro pointz			; Load a pointer into Z-register. Z = 1st SRAM_adrs + <2nd SRAM_adrs>
lds zl,@1				; The first parameter is the starting address, e.g., of an array
ldi zh,0				; The 2nd address points to an SRAM location containing the displacement value
ldi temp,low(@0)		; SRAM_adrs, SRAM_adrs
add zl,temp
ldi temp,high(@0)
adc zh,temp
.endmacro

.macro textadrs			;Load an immediate value into the 16-bit register Z
						; so it can be used as a byte address in LPM instructions 
ldi zl, low(@0*2)		; 16-bit-literal (string adrs in pgm memory)
ldi zh, high(@0*2)
.endmacro


; **********************************************************************************************

; 					26.  Save and restore register contents  ***********



.macro pushf			; Push the flag register (SREG) onto stack
in temp,sreg
push temp
.endmacro

.macro popf				; Retrieve the flag register (SREG) from stack
pop temp
out sreg,temp
.endmacro

.macro pushft			; Save the flag register (SREG) and the register TEMP onto stack
push temp
in temp,sreg
push temp
.endmacro

.macro popft			; Retrieve the flag register (SREG) and the register TEMP from stack
pop temp
out sreg,temp
pop temp
.endmacro

.macro pushft10			; Save the flag register (SREG), the register TEMP, and the registers r0, r1 onto stack
push temp
in temp,sreg
push temp
push r1
push r0
.endmacro

.macro popft10			; Retrieve the flag register (SREG), the register TEMP, and the registers r0, r1 from Stack
pop temp
out sreg,temp
pop temp
pop r0
pop r1
.endmacro

.macro pushaf			; Save the flag register (SREG), the register TEMP, and the A-register onto stack
push temp
in temp,sreg
push temp
push r25
push r24
.endmacro

.macro popaf			; Retrieve the flag register (SREG), the register TEMP, and the A-register from stack
pop r24
pop r25
pop temp
out sreg,temp
pop temp
.endmacro

.macro pushaf10			; Save the flag register (SREG), the registers TEMP, r0, r1 and the A-register onto stack
push temp
in temp,sreg
push temp
push r1  
push r0
push r25
push r24
.endmacro

.macro popaf10			; Retrieve the flag register (SREG), the registers TEMP, r0, r1 and the A-register from stack
pop r24
pop r25
pop r0
pop r1
pop temp
out sreg,temp
pop temp
.endmacro

.macro pushregs			; Save all registres belonging to the register model onto stack
push temp				; Flags (SREG), TEMP, r0, r1, A, B, C, X, Y, Z
in temp,sreg
push temp
push r1
push r0
push ah
push al
push bh
push bl
push ch
push cl
push xh
push xl
push yh
push yl
push zh
push zl
.endmacro

.macro popregs			; Retrieve all registres belonging to the register model from stack
pop zl						; Flags (SREG), TEMP, r0, r1, A, B, C, X, Y, Z
pop zh
pop yl
pop yh
pop xl
pop xh
pop cl
pop ch
pop bl
pop bh
pop al
pop ah
pop r0
pop r1
pop temp
out sreg,temp
pop temp
.endmacro



; **********************************************************************************************

; 								26.  Stack-machine emulation  **********************************


.macro pusha			; Push the content of the A-register onto the stack (16 bits)
push ah
push al
.endmacro

.macro pushb			; Push the content of the B-register onto the stack (16 bits)
push r23
push r22
.endmacro

.macro pushc			; Push the content of the C-register onto the stack (16 bits)
push r21
push r20
.endmacro

.macro pushx			; Push the content of the X-register onto the stack (16 bits)
push r27
push r26
.endmacro

.macro pushy			; Push the content of the Y-register onto the stack (16 bits)
push r29
push r28
.endmacro

.macro pushz			; Push the content of the Z-register onto the stack (16 bits)
push r31
push r30
.endmacro

.macro popa 			; Transfer TOS content into the A-register
pop al
pop ah
.endmacro

.macro popb				; Transfer TOS content into the B-register
pop r22
pop r23
.endmacro

.macro popc				; Transfer TOS content into the C-register
pop r20
pop r21
.endmacro

.macro popx				; Transfer TOS content into the X-register
pop r26
pop r27
.endmacro

.macro popy				; Transfer TOS content into the Y-register
pop r28
pop r29
.endmacro

.macro popz				; Transfer TOS content into the Z-register
pop r30
pop r31
.endmacro

.macro pushlit			; Push an 8-bit literal  onto the stack 
ldi temp,@0				; literal
push temp
.endmacro

.macro pushwlit			; Push an 16-bit literal  onto the stack 
ldi temp,high(@0)		; 16-bit-literal
push temp
ldi temp,low(@0)
push temp
.endmacro

.macro fetch			; Replace the address on TOS with the addressed byte
popx					
ld temp,x+				; The X-register contains the next memory address
push temp
.endmacro

.macro fetchw			; Replace the address on TOS with the addressed 16-bit word
popx					
ld r0,x+				; The X-register contains the next memory address
ld r1,x+
push r1
push r0
.endmacro

.macro store			; Store the byte on TOS at the address on TOS+1
pop temp				; Remove byte and address from stack				
popx
st x+,temp				; The X-register contains the next memory address
.endmacro

.macro storew			; Store the word on TOS at the address on TOS+2
pop r0					; Remove word and address from stack
pop r1
popx
st x+,r0				; The X-register contains the next memory address
st x+,r1
.endmacro

.macro drop				; Remove the byte from TOS
pop temp
.endmacro

.macro dropw			; Remove the word from TOS
pop temp
pop temp
.endmacro

.macro swaps			; Swap both bytes on TOS
pop temp
pop al
push temp
push al					; The AL-register contains the new TOS byte
.endmacro

.macro swapsw			; Swap both words on TOS
pop r0
pop r1
pop al
pop ah
push r1
push r0
push ah					; The A-register contains the new TOS word
push al
.endmacro

.macro dup				; Duplicate the byte on TOS 
pop al
push al
push al					; The AL-register contains the duplicted TOS byte
.endmacro

.macro dupw				; Duplicate the word on TOS 
popa
pusha
pusha
.endmacro

.macro over				; Duplicate the byte on TOS 
pop temp				; byte on TOS
pop al					; byte on TOS + 1
push al					; The AL-register contains the byte from TOS+1 inserted on TOS
push temp
.endmacro

.macro overw			; Duplicate the word on TOS 
pop r0					; word on TOS
pop r1				
popa					; word on TOS + 1
pusha					; The AL-register contains the word from TOS+1 inserted on TOS
push r1
push r0
pusha					; The A-register contains the word from TOS+1 inserted on TOS
.endmacro

.macro fetchr			; Transfer TOS content into the X-register and load 8-bit register
popx					; with the addressed byte
ld @0,x+				; The X-register contains the next memory address
.endmacro

.macro fetchwa			; Transfer TOS content into the X-register and load the A-register
popx					; with the addressed word
ld a,x+
ld ah,x+				; The X-register contains the next memory address
.endmacro

.macro fetchwb			; Transfer TOS content into the X-register and load the B-register
popx					; with the addressed word
ld b,x+					; The X-register contains the next memory address
ld bh,x+
.endmacro

.macro fetchwc			; Transfer TOS content into the X-register and load the C-register
popx					; with the addressed word
ld c,x+					; The X-register contains the next memory address
ld ch,x+
.endmacro

.macro fetchwy			; Transfer TOS content into the X-register and load the Y-register
popx					; with the addressed word
ld yl,x+				; The X-register contains the next memory address
ld yh,x+
.endmacro

.macro fetchwz			; Transfer TOS content into the X-register and load the Z-register
popx					; with the addressed word
ld zl,x+				; The X-register contains the next memory address
ld zh,x+
.endmacro

.macro storr			; Transfer TOS content into the X-register and store  
popx					; the content of the 8-bit register on the adddress in X
st x+,@0				; The X-register contains the next memory address
.endmacro

.macro storwa			; Transfer TOS content into the X-register and store the word
popx					; in the A-register on the adddress in X
st x+,al				; The X-register contains the next memory address
st x+,ah
.endmacro

.macro storwb			; Transfer TOS content into the X-register and store the word
popx					; in the B-register on the adddress in X
st x+,bl				; The X-register contains the next memory address
st x+,bh
.endmacro

.macro storwc			; Transfer TOS content into the X-register and store the word 
popx					; in the C-register on the adddress in X
st x+,cl				; The X-register contains the next memory address
st x+,ch
.endmacro

.macro pusheab			; Push an effective address onto the stack 
lds r0,@0				; 1st parameter = base ddress in SRAM
lds r1,@0+1				; 2nd parameter = displacement literal
ldi temp,low(@1)
add r0,temp
ldi temp,high(@1)
adc r1,temp
push r1
push r0
.endmacro

.macro pusheabd			; Push an effective address onto the stack 
lds r0,@0				; 1st parameter = base ddress in SRAM
lds r1,@0+1				; 2nd parameter = displacement address
lds temp,@1
add r0,temp
lds temp,@1+1
adc r1,temp
push r1
push r0
.endmacro

.macro pusheax			; Push an effective address onto the stack 
ldi temp,low(@0)		; Effective adrs := <X> + 16-bit literal. 
add xl,temp
ldi temp,high(@0)
adc xh,temp
pushx
.endmacro

.macro pusheay			; Push an effective address onto the stack 
ldi temp,low(@0)		; Effective adrs := <Y> + 16-bit literal. 
add yl,temp
ldi temp,high(@0)
adc yh,temp
pushy
.endmacro

.macro pusheaz			; Push an effective address onto the stack 
ldi temp,low(@0)		; Effective adrs := <Z> + 16-bit literal. 
add zl,temp
ldi temp,high(@0)
adc zh,temp
pushy
.endmacro


; **********************************************************************************************

; 					27.  Subroutine call and parameter addressing  ***********


.macro stackindexx		; In the X-register, an address pointer is provided
						; pointing to the first address above the saved return address	
						; in the stack. This allows to address parameters, which have been pushed 
						; prior to the subroutine call
in temp,sreg				; save interrupt flag 
cli
in xl,spl					; fetch the stack pointer
in xh,sph
adiw xl,3					; make adrs point to the uppermost parameter byte
out sreg,temp				; restore interrupt flag
.endmacro

.macro clrstackretx		; Return from a subroutine called with parameters on the stack
						; The subroutine must begin with a stackindexx macro
						; clrstackretx	is to be programmed instead of a ret
						; The X-register must have the content as established by the previous stackindexx macro
						; Parameter = Number of bytes to be removed from the stack
						; SP in X, return address in Z. Macro uses register r0
in r0,sreg					; save interrupt flag 
cli
pop zh						; SP points to the return adress
pop zl						; Return adrs to Z 
ldi temp,low(@0-1)			; Incremented SP
add xl,temp
ldi temp,high(@0-1)
adc xh,temp
out spl,xl					; restore modified stack pointer			
out sph,xh
out sreg,r0					; restore interrupt flag
ijmp						; Return with instruction adrs in Z
.endmacro

.macro stackindexy		; In the Y-register, an address pointer is provided
						; pointing to the first address beneath the saved return address	
						; in the stack. This allows to address parameters, which have been pushed 
						; prior to the subroutine call
in temp,sreg				; save interrupt flag 
cli
in yl,spl					; fetch the stack pointer
in yh,sph
adiw yl,3					; make adrs point to the uppermost parameter byte
out sreg,temp				; restore interrupt flag
.endmacro

.macro clrstackrety		; Return from a subroutine called with parameters on the stack
						; The subroutine must begin with a stackindexy macro
						; clrstackretx	is to be programmed instead of a ret
						; The Y-register must have the content as established by the previous stackindexx macro
						; Parameter = Number of bytes to be removed from the stack
						; SP in Y, return address in Z. Macro uses register r0
in r0,sreg					; save interrupt flag 
cli
pop zh						; SP points to the return adress
pop zl						; Return adrs to Z 
ldi temp,low(@0-1)			; Incremented SP
add yl,temp
ldi temp,high(@0-1)
adc yh,temp
out spl,yl					; restore modified stack pointer			
out sph,yh
out sreg,r0					; restore interrupt flag
ijmp						; Return with instruction adrs in Z
.endmacro

.macro enter			; C-like industry standard function call. The Y-register serves as Frame Pointer (FP)
						; Parameter = Number of bytes to be reserved for local variables
in temp,sreg				; save interrupt flag 
cli
push yh						; save frame pointer onto stack
push yl
in yl,spl					; make the stack pointer to the new frame pointer
in yh,sph
movw r0,yl					; save the new frame pointer
subi yl,low(@0)				; decrement the new stack pointer to make room for local variables
sbci yh,high(@0)
out spl,yl					; establish the new stack pointer
out sph,yh
movw yl,r0					; restore the new frame pointer
out sreg,temp				; restore interrupt flag
.endmacro

.macro leave			; C-like industry standard function return. The Y-register serves as Frame Pointer (FP)
in temp,sreg				; save interrupt flag
cli
out spl,yl					; restore the old stack pointer (FP => SP)
out sph,yh
pop yl						; restore the old frame pointer out of the stack
pop yh
out sreg,temp				; restore interrupt flag
.endmacro

.macro AVR_enter			; C-like function call adapted to the AVR architecture
							; The Y-register serves as Frame Pointer (FP)
							; Parameter = Number of bytes to be reserved for local variables
in temp,sreg				; save interrupt flag 
cli
push yh						; save frame pointer onto stack
push yl
in yl,spl					; read the stack pointer
in yh,sph
subi yl,low(@0)				; decrement the stack pointer to make room for local variables
sbci yh,high(@0)
out spl,yl					; establish the new stack pointer and frame pointer. Both are equal
out sph,yh
out sreg,temp				; restore interrupt flag
.endmacro

.macro AVR_Leave		; C-like function return adapted to the AVR architecture
						; The Y-register serves as Frame Pointer (FP)
in temp,sreg				; save interrupt flag
cli
subi yl,low(-@0)			; increment by the number Number of bytes to be reserved for local variables
sbci yh,high(-@0)
out spl,yl					; restore the old stack pointer (FP => SP)
out sph,yh
pop yl						; restore the old frame pointer out of the stack
pop yh
out sreg,temp				; restore interrupt flag
.endmacro





